# -*- coding: utf-8 -*-

"""
 Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# pylint: disable=invalid-name

import sqlite3

from typing import Any, List, Optional, TYPE_CHECKING

from mi18n import i18nc

if TYPE_CHECKING:
    from tile import Tile
    from board import WindDisc



class Wind:
    """we want to use an wind for indices.

    char is the wind as a char (native string)
    """

    char: str
    discSvgName: str
    all : List['Wind'] = []
    all4 : List['Wind'] = []

    tile : 'Tile'
    disc:'WindDisc'

    # X is used in scoring games for moving back a tile to the central court
    eswnx = 'ESWNX'
    eswnx_i18n = 'ESWNX'

    def __new__(cls, *args:Any) ->'Wind':
        if not Wind.all:
            Wind.all = [object.__new__(cls) for cls in (_East, _South, _West, _North, _NoWind)]
            Wind.all4 = list(Wind.all[:4])
            Wind.eswnx_i18n = i18nc('kajongg:keyboard commands for moving tiles to the players '
                             'with wind ESWN or to the central tile selector (X)', Wind.eswnx)
        if len(args) == 1:
            windIdent = args[0]
            assert cls is Wind, f'{cls.__name__}({windIdent}) is illegal'
            windIdx = 'eswn'.index(windIdent.lower())
            return Wind.all[windIdx]
        assert not args and cls is not Wind, 'Wind() must have exactly one argument'

        for result in Wind.all:
            if isinstance(result, cls):
                # return the correct Wind subclass like _East
                return result
        assert False

    def __eq__(self, other:Any) ->bool:
        if not other:
            return False
        if isinstance(other, self.__class__):
            return True
        if isinstance(other, Wind):
            return False
        try:
            return str(self.char) == other.upper()
        except AttributeError:
            return False

    def __gt__(self, other:Any) ->bool:
        assert isinstance(other, Wind)
        return self.__index__() > other.__index__()

    def __lt__(self, other:Any) ->bool:
        assert isinstance(other, Wind)
        return self.__index__() < other.__index__()

    def __ge__(self, other:Any) ->bool:
        assert isinstance(other, Wind)
        return self.__index__() >= other.__index__()

    def __le__(self, other:Any) ->bool:
        assert isinstance(other, Wind)
        return self.__index__() <= other.__index__()

    def __hash__(self) ->int:
        return self.__index__()

    def __index__(self) ->int:
        raise NotImplementedError

    def __str__(self) ->str:
        return self.char

    def __repr__(self) ->str:
        return f'Wind.{self.char}'

    def __next__(self) ->'Wind':
        """after North, return NoWind"""
        return Wind.all[self.__index__() +1]

    def __conform__(self, protocol) ->Optional[str]:
        if protocol is sqlite3.PrepareProtocol:
            return self.char
        return None

    @classmethod
    def normalized(cls, key:int) -> Optional['Wind']:
        """translate i18n key to wanted Wind. None if key is not a Wind character"""
        try:
            return Wind.all[cls.eswnx_i18n.index(chr(key))]
        except ValueError:
            return None

class _East(Wind):
    """East"""
    char = 'E'
    discSvgName = 'g4657'  # WIND_2 etc have a border

    def __index__(self) ->int:
        return 0


class _South(Wind):
    """South"""
    char = 'S'
    discSvgName = 'g3980'

    def __index__(self) ->int:
        return 1


class _West(Wind):
    """West"""
    char = 'W'
    discSvgName = 'g3192'

    def __index__(self) ->int:
        return 2


class _North(Wind):
    """North"""
    char = 'N'
    discSvgName = 'g4290'

    def __index__(self) ->int:
        return 3


class _NoWind(Wind):
    """no wind"""
    char = 'X'

    def __index__(self) ->int:
        return 4


East = _East()
South = _South()
West = _West()
North = _North()
NoWind = _NoWind()

def convert_from_sql(raw:bytes) -> 'Wind':
    """see https://docs.python.org/3/library/sqlite3.html"""
    return Wind(raw.decode())


sqlite3.register_converter("wind", convert_from_sql)



#!/usr/bin/env python3

"""
Copyright (c) 2010,2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>
Redistribution and use is allowed according to the terms of the BSD license.
For details see the accompanying COPYING-CMAKE-SCRIPTS file.
"""

import sys

try:
    from twisted.spread import pb
    _ = pb
except BaseException:
    print('twisted_version:{}'.format('0.0.0'))
    sys.exit(0)
from twisted import __version__
print('twisted_version:{}'.format(__version__))



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# pylint: disable=wrong-import-order

import sys
from itertools import chain

import logging
from typing import Any, TYPE_CHECKING, Optional, Union, Type, cast

from log import logError, logDebug
from common import Options, Internal, isAlive, Debug, handleSignals

NOTFOUND = []

try:
    from qt import Qt, QEvent, QMetaObject, QTimer
    from qt import QWidget, QGridLayout, QAction
except ImportError as importError:
    NOTFOUND.append(f'Please install PyQt6 or Pyside6: {importError}')

try:
    from twisted.spread import pb # pylint: disable=unused-import
    from twisted.internet.error import ReactorNotRunning
except ImportError as importError:
    NOTFOUND.append(f'Package python3-twisted is missing or too old (I need 16.6.0): {importError}')


try:
    from mi18n import i18n, i18nc
    from kde import KXmlGuiWindow, KStandardAction

    from board import FittingView
    from playerlist import PlayerList
    from tileset import Tileset
    from background import Background
    from scoring import scoreGame
    from scoringdialog import ScoreTable, ExplainView
    from humanclient import HumanClient
    from rulesetselector import RulesetSelector
    from animation import afterQueuedAnimations, AnimationSpeed
    from chat import ChatWindow
    from scene import PlayingScene, ScoringScene
    from configdialog import ConfigDialog
    from statesaver import StateSaver
    from util import checkMemory
    from kdestub import Action, KApplication

except ImportError as importError:
    NOTFOUND.append(f'Kajongg is not correctly installed: modules: {importError}')

if NOTFOUND:
    logError("\n".join(f" * {s}" for s in NOTFOUND), showStack=False)
    sys.exit(3)

if TYPE_CHECKING:
    from deferredutil import Request
    from twisted.internet.defer import Deferred
    from qt import QSize
    from scene import GameScene

def cleanExit(*unusedArgs:Any) ->None:
    """close sqlite3 files before quitting"""
    if isAlive(Internal.mainWindow):
        if Debug.quit:
            logDebug('cleanExit calling mainWindow.close')
        assert Internal.mainWindow  # for mypy
        Internal.mainWindow.close()
    else:
        # this must be very early or very late
        if Debug.quit:
            logDebug('cleanExit calling sys.exit(0)')
        # sys.exit(0)
        MainWindow.aboutToQuit()

handleSignals(cleanExit)


class MainWindow(KXmlGuiWindow):

    """the main window"""
    # pylint: disable=too-many-instance-attributes

    def __init__(self) ->None:
        # see https://marc.info/?l=kde-games-devel&m=120071267328984&w=2
        super().__init__()
        Internal.app.aboutToQuit.connect(self.aboutToQuit)
        self.exitConfirmed:Optional[bool] = None
        self.exitReady:Optional[bool] = None
        self.exitWaitTime:Optional[int] = None
        Internal.mainWindow = self
        self._scene:Optional['GameScene'] = None
        self.centralView: FittingView
        self.background:Background = Background()
        self.playerWindow:Optional[PlayerList] = None
        self.rulesetWindow:Optional[RulesetSelector] = None
        self.confDialog:Optional[ConfigDialog] = None
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.__installReactor()
        if Options.gui:
            KStandardAction.preferences(
                self.showSettings,
                self.actionCollection())
            self.setupUi()
            self.setupGUI()
            assert Internal.Preferences
            Internal.Preferences.addWatch(
                'tilesetName',
                self.tilesetNameChanged)
            Internal.Preferences.addWatch(
                'backgroundName',
                self.backgroundChanged)
            self.retranslateUi()
            for action in self.toolBar().actions():
                if 'onfigure' in action.text():
                    action.setPriority(QAction.Priority.LowPriority)
            if Options.host and not Options.demo:
                self.scene = PlayingScene(self)
                HumanClient()
            StateSaver(self)
            self.show()
        else:
            HumanClient()

    @staticmethod
    def __installReactor() ->None:
        """install the twisted reactor"""
        if not hasattr(Internal, 'reactor'):
            import qtreactor
            qtreactor.install()
            from twisted.internet.interfaces import IReactorCore
            from twisted.internet import reactor as reactor_module
            reactor = cast(IReactorCore, reactor_module)
            reactor.runReturn(installSignalHandlers=False)
            Internal.reactor = reactor
            if Debug.quit:
                logDebug('Installed qtreactor')

    @property
    def scene(self) ->Optional['GameScene']:
        """a proxy"""
        return self._scene

    @scene.setter
    def scene(self, value:Optional['GameScene']) ->None:
        """if changing, updateGUI"""
        if not isAlive(self):
            return
        if self._scene == value:
            return
        if not value:
            self.actionChat.setChecked(False)
            self.actionExplain.setChecked(False)
            self.actionScoreTable.setChecked(False)
            self.actionExplain.setData(ExplainView)
            self.actionScoreTable.setData(ScoreTable)
        self._scene = value
        # contrary to what the Qt doc says, scene can be set to None
        self.centralView.setScene(value)  # type:ignore[arg-type]
        self.adjustMainView()
        self.updateGUI()
        canDemo = not value or isinstance(value, PlayingScene)
        self.actionChat.setEnabled(canDemo)
        self.actionAutoPlay.setEnabled(canDemo)
        self.actionExplain.setEnabled(value is not None)
        self.actionScoreTable.setEnabled(value is not None)

    def sizeHint(self) ->'QSize':
        """give the main window a sensible default size"""
        result = super().sizeHint()
        result.setWidth(result.height() * 3 // 2)
                        # we want space to the right for the buttons
        # the default is too small. Use at least 2/3 of screen height and 1/2
        # of screen width:
        available = KApplication.desktopSize()
        height = max(result.height(), available.height() * 2 // 3)
        width = max(result.width(), available.width() // 2)
        result.setHeight(height)
        result.setWidth(width)
        return result

    def _kajonggToggleAction(self, name:str, icon:str, shortcut:Optional[str]=None,
        actionData:Optional[Union[str, Type['QWidget']]]=None) ->Action:
        """a checkable action"""
        res = Action(self,
            name,
            icon,
            shortcut=shortcut,
            actionData=actionData)
        res.setCheckable(True)
        if actionData is not None:
            res.toggled.connect(self._toggleWidget)
        return res

    def setupUi(self) ->None:
        """create all other widgets
        we could make the scene view the central widget but I did
        not figure out how to correctly draw the background with
        QGraphicsView/QGraphicsScene.
        QGraphicsView.drawBackground always wants a pixmap
        for a huge rect like 4000x3000 where my screen only has
        1920x1200"""
        self.setObjectName("MainWindow")
        centralWidget = QWidget()
        self.centralView = FittingView()
        layout = QGridLayout(centralWidget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.centralView)
        self.setCentralWidget(centralWidget)
        self.centralView.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        assert Internal.Preferences
        _ = Internal.Preferences.windTilesetName
        assert isinstance(_, str)
        self.windTileset = Tileset(_)
        self.adjustMainView()
        self.actionScoreGame = Action(
            self,
            "scoreGame",
            "draw-freehand",
            self.startScoringGame,
            "Ctrl+C")
        self.actionPlayGame = Action(
            self,
            "play",
            "arrow-right",
            self.playGame,
            "Ctrl+N")
        self.actionAbortGame = Action(
            self,
            "abort",
            "dialog-close",
            self.abortAction,
            "Ctrl+W")
        self.actionAbortGame.setEnabled(False)
        self.actionQuit = Action(
            self,
            "quit",
            "application-exit",
            self.close,
            "Ctrl+Q")
        self.actionPlayers = Action(
            self,
            "players", "im-user", self.slotPlayers)
        self.actionRulesets = Action(
            self,
            "rulesets",
            "games-kajongg-law",
            self.slotRulesets)
        self.actionChat = self._kajonggToggleAction("chat", "call-start",
                                                    shortcut="Ctrl+H", actionData=ChatWindow)
        self.actionChat.setEnabled(False)
        self.actionAngle = Action(
            self,
            "angle",
            "object-rotate-left",
            self.changeAngle,
            "Ctrl+G")
        self.actionAngle.setEnabled(False)
        self.actionScoreTable = self._kajonggToggleAction(
            "scoreTable", "format-list-ordered",
            "Ctrl+T", actionData=ScoreTable)
        self.actionScoreTable.setEnabled(False)
        self.actionExplain = self._kajonggToggleAction(
            "explain", "applications-education",
            "Ctrl+E", actionData=ExplainView)
        self.actionExplain.setEnabled(False)
        self.actionFullscreen = self._kajonggToggleAction(
            "fullscreen", "view-fullscreen",
            shortcut="Ctrl+Shift+F")
        self.actionFullscreen.toggled.connect(self.fullScreen)
        self.actionAutoPlay = Action(
            self,
            "demoMode",
            "arrow-right-double",
            None,
            "Ctrl+D")
        self.actionAutoPlay.setCheckable(True)
        self.actionAutoPlay.setEnabled(True)
        self.actionAutoPlay.toggled.connect(self._toggleDemoMode)
        self.actionAutoPlay.setChecked(Internal.autoPlay)
        QMetaObject.connectSlotsByName(self)

    def playGame(self) ->None:
        """manual wish for a new game"""
        if not Internal.autoPlay:
            # only if no demo game is running
            self.startPlayingGame()

    def startPlayingGame(self) ->None:
        """play a computer game: log into a server and show its tables"""
        self.scene = PlayingScene(self)
        HumanClient()

    def startScoringGame(self) ->None:
        """start a scoring scene"""
        scene = ScoringScene(self)
        game = scoreGame()
        if game:
            self.scene = scene
            scene.game = game
            game.throwDices()
            self.updateGUI()

    def fullScreen(self, toggle:bool) ->None:
        """toggle between full screen and normal view"""
        _ = self.windowState()
        if toggle:
            new_state = _ | Qt.WindowState.WindowFullScreen
        else:
            new_state = _ & ~Qt.WindowState.WindowFullScreen
        self.setWindowState(cast(Qt.WindowState, new_state))

    def closeMe(self) ->None:
        """for QTimer"""
        self.close()

    def close(self, unusedResult:Any=None) ->bool:
        """wrap close() because we call it with a QTimer"""
        if isAlive(self):
            return super().close()
        return True  # is closed

    def closeEvent(self, event:Optional[QEvent]) ->None:
        super().closeEvent(event)
        if event:
            if event.isAccepted() and self.exitReady:
                QTimer.singleShot(5000, self.aboutToQuit)

    def queryClose(self) ->bool:
        """queryClose, queryExit and aboutToQuit are no
        ideal match for the async Deferred approach.

        At app start, self.exitConfirmed and exitReady are None.

        queryClose will show a confirmation prompt if needed, but
        it will not wait for the answer. queryClose always returns True.

        Later, when the user confirms exit, self.exitConfirmed will be set.
        If the user cancels exit, self.exitConfirmed = False, otherwise
        self.close() is called. This time, no prompt will appear because the
        game has already been aborted.

        queryExit will return False if exitConfirmed or exitReady are not True.
        Otherwise, queryExit will set exitReady to False and asynchronously start
        shutdown. After the reactor stops running, exitReady is set to True,
        and self.close() is called. This time it should fall through everywhere,
        having queryClose() and queryExit() return True.

        and it will reset exitConfirmed to None.

        Or in other words: If queryClose or queryExit find something that they
        should do async like asking the user for confirmation or terminating
        the client/server connection, they start async operation and append
        a callback which will call self.close() when the async operation is
        done. This repeats until queryClose() and queryExit() find nothing
        more to do async. At that point queryExit says True
        and we really end the program.
        """

        def confirmed(result:Any) ->None:
            """quit if the active game has been aborted"""
            self.exitConfirmed = bool(result)
            if Debug.quit:
                if self.exitConfirmed:
                    logDebug('mainWindow.queryClose confirmed')
                else:
                    logDebug('mainWindow.queryClose not confirmed')
            # start closing again. This time no question will appear, the game
            # is already aborted
            if self.exitConfirmed:
                assert isAlive(self)
                self.close()
            else:
                self.exitConfirmed = None

        def cancelled(result:Any) ->None:
            """just do nothing"""
            if Debug.quit:
                logDebug(f'mainWindow.queryClose.cancelled: {result}')
            self.exitConfirmed = None
        if self.exitConfirmed is False:
            # user is currently being asked
            return False
        if self.exitConfirmed is None:
            if self.scene:
                self.exitConfirmed = False
                self.abortAction().addCallbacks(confirmed, cancelled)
            else:
                self.exitConfirmed = True
                if Debug.quit:
                    logDebug(
                        'MainWindow.queryClose not asking, exitConfirmed=True')
        return True

    def queryExit(self) ->bool:
        """see queryClose"""
        def quitDebug(msg:str) ->None:
            """reducing branches in queryExit"""
            if Debug.quit:
                logDebug(msg)

        if self.exitReady:
            quitDebug('MainWindow.queryExit returns True because exitReady is set')
            return True
        if self.exitConfirmed:
            # now we can get serious
            self.exitReady = False
            for widget in chain(
                    (x.tableList for x in HumanClient.humanClients), [
                        self.confDialog,
                        self.rulesetWindow, self.playerWindow]):
                if isAlive(widget):
                    assert widget
                    widget.hide()
            if self.exitWaitTime is None:
                self.exitWaitTime = 0
            if hasattr(Internal, 'reactor') and Internal.reactor.running:
                self.exitWaitTime += 10
                if self.exitWaitTime % 1000 == 0:
                    logDebug(
                        f'waiting since {int(self.exitWaitTime // 1000)} seconds for reactor to stop')
                try:
                    quitDebug('now stopping reactor')
                    Internal.reactor.stop()  # type:ignore[misc]
                    assert isAlive(self)
                    QTimer.singleShot(10, self.closeMe)
                except ReactorNotRunning:
                    self.exitReady = True
                    quitDebug(
                        'MainWindow.queryExit returns True: It got exception ReactorNotRunning')
            else:
                self.exitReady = True
                quitDebug('MainWindow.queryExit returns True: Reactor is not running')
        return bool(self.exitReady)

    @staticmethod
    def aboutToQuit() ->None:
        """now all connections to servers are cleanly closed"""
        mainWindow = Internal.mainWindow
        Internal.mainWindow = None
        if mainWindow:
            if Debug.quit:
                logDebug('aboutToQuit starting')
            if mainWindow.exitWaitTime is not None:
                if mainWindow.exitWaitTime > 1000.0 or Debug.quit:
                    logDebug(
                        f'reactor stopped after {int(mainWindow.exitWaitTime)} ms')
        else:
            if Debug.quit:
                logDebug('aboutToQuit: mainWindow is already None')
            # this does not happen with PyQt6, only with PySide6
            # last test was with Pyside6.7.2
            # return here to avoid recursion in StateSaver
            return
        StateSaver.saveAll()
        Internal.app.quit()
        try:
            # if we are killed while loading, Internal.db may not yet be
            # defined
            if Internal.db:
                Internal.db.close()
        except NameError:
            pass
        checkMemory()
        logging.shutdown()
        if Debug.quit:
            logDebug('aboutToQuit ending')

    def abortAction(self) ->'Deferred':
        """abort current game"""
        if Debug.quit:
            logDebug('mainWindow.abortAction invoked')
        assert self.scene
        return self.scene.abort()

    def retranslateUi(self) ->None:
        """retranslate"""
        self.actionScoreGame.setText(
            i18nc('@action:inmenu', "&Score Manual Game"))
        self.actionScoreGame.setIconText(
            i18nc('@action:intoolbar', 'Manual Game'))
        self.actionScoreGame.setWhatsThis(
            i18nc('kajongg @info:tooltip',
                  '&Score a manual game.'))

        self.actionPlayGame.setText(i18nc('@action:intoolbar', "&Play"))
        self.actionPlayGame.setPriority(QAction.Priority.LowPriority)
        self.actionPlayGame.setWhatsThis(
            i18nc('kajongg @info:tooltip', 'Start a new game.'))

        self.actionAbortGame.setText(i18nc('@action:inmenu', "&Abort Game"))
        self.actionAbortGame.setPriority(QAction.Priority.LowPriority)
        self.actionAbortGame.setWhatsThis(
            i18nc('kajongg @info:tooltip',
                  'Abort the current game.'))

        self.actionQuit.setText(i18nc('@action:inmenu', "&Quit Kajongg"))
        self.actionQuit.setPriority(QAction.Priority.LowPriority)

        self.actionPlayers.setText(i18nc('@action:intoolbar', "&Players"))
        self.actionPlayers.setWhatsThis(
            i18nc('kajongg @info:tooltip',
                  'define your players.'))

        self.actionRulesets.setText(i18nc('@action:intoolbar', "&Rulesets"))
        self.actionRulesets.setWhatsThis(
            i18nc('kajongg @info:tooltip',
                  'customize rulesets.'))

        self.actionAngle.setText(
            i18nc('@action:inmenu',
                  "&Change Visual Angle"))
        self.actionAngle.setIconText(i18nc('@action:intoolbar', "Angle"))
        self.actionAngle.setWhatsThis(
            i18nc('kajongg @info:tooltip',
                  "Change the visual appearance of the tiles."))

        self.actionFullscreen.setText(
            i18nc('@action:inmenu',
                  "F&ull Screen Mode"))

        self.actionScoreTable.setText(
            i18nc('kajongg @action:inmenu', "&Score Table"))
        self.actionScoreTable.setIconText(
            i18nc('kajongg @action:intoolbar', "&Scores"))
        self.actionScoreTable.setWhatsThis(i18nc('kajongg @info:tooltip',
                                                 "Show or hide the score table for the current game."))

        self.actionExplain.setText(i18nc('@action:inmenu', "&Explain Scores"))
        self.actionExplain.setIconText(i18nc('@action:intoolbar', "&Explain"))
        self.actionExplain.setWhatsThis(i18nc('kajongg @info:tooltip',
                                              'Explain the scoring for all players in the current game.'))

        self.actionAutoPlay.setText(i18nc('@action:inmenu', "&Demo Mode"))
        self.actionAutoPlay.setPriority(QAction.Priority.LowPriority)
        self.actionAutoPlay.setWhatsThis(i18nc('kajongg @info:tooltip',
                                               'Let the computer take over for you. Start a new local game if needed.'))

        self.actionChat.setText(i18n("C&hat"))
        self.actionChat.setWhatsThis(
            i18nc('kajongg @info:tooltip',
                  'Chat with the other players.'))

    def changeEvent(self, event:Optional[QEvent]) ->None:
        """when the applicationwide language changes, recreate GUI"""
        if event:
            super().changeEvent(event)
            if event.type() == QEvent.Type.LanguageChange:
                self.setupGUI()
                self.retranslateUi()

    def slotPlayers(self) ->None:
        """show the player list"""
        if not self.playerWindow:
            self.playerWindow = PlayerList(self)
        self.playerWindow.show()

    def slotRulesets(self) ->None:
        """show the player list"""
        if not self.rulesetWindow:
            self.rulesetWindow = RulesetSelector()
        self.rulesetWindow.show()

    def adjustMainView(self) ->None:
        """adjust the view such that exactly the wanted things are displayed
        without having to scroll"""
        if not isAlive(self):
            return
        view, scene = self.centralView, self.scene
        if scene:
            scene.adjustSceneView()
            view.fitInView(scene.itemsBoundingRect(), Qt.AspectRatioMode.KeepAspectRatio)

    @afterQueuedAnimations  # type:ignore[arg-type]
    def backgroundChanged(self, unusedDeferredResult:'Deferred', unusedOldName:str, newName:str) ->None:
        """if the wanted background changed, apply the change now"""
        centralWidget = self.centralWidget()
        if centralWidget:
            self.background = Background(newName)
            self.background.setPalette(centralWidget)
            centralWidget.setAutoFillBackground(True)

    @afterQueuedAnimations  # type:ignore[arg-type]
    def tilesetNameChanged(
            self, unusedDeferredResult:'Deferred', unusedOldValue:Optional[str]=None,
            unusedNewValue:Optional[str]=None) ->None:
        """if the wanted tileset changed, apply the change now"""
        if self.centralView:
            with AnimationSpeed():
                if self.scene:
                    self.scene.applySettings()
            self.adjustMainView()

    @afterQueuedAnimations  # type:ignore[arg-type]
    def showSettings(self, unusedDeferredResult:'Request', unusedChecked:Optional[bool]=None) ->None:
        """show preferences dialog. If it already is visible, do nothing"""
        # This is called by the triggered() signal. So why does KDE
        # not return the bool checked?
        if ConfigDialog.showDialog("settings"):
            return
        # if an animation is running, Qt segfaults somewhere deep
        # in the SVG renderer rendering the wind tiles for the tile
        # preview
        self.confDialog = ConfigDialog(self, "settings")
        self.confDialog.show()

    def _toggleWidget(self, checked:bool) ->None:
        """user has toggled widget visibility with an action"""
        assert self.scene
        action = self.sender()
        assert isinstance(action, Action), action
        actionData = action.data()
        if checked:
            if isinstance(actionData, type):
                clsName = actionData.__name__
                actionData = actionData(scene=self.scene)
                action.setData(actionData)
                setattr(
                    self.scene,
                    clsName[0].lower() + clsName[1:],
                    actionData)
            actionData.show()
            actionData.raise_()
        else:
            assert actionData
            actionData.hide()

    def _toggleDemoMode(self, checked:bool) ->None:
        """switch on / off for autoPlay"""
        if self.scene:
            assert isinstance(self.scene, PlayingScene)
            self.scene.toggleDemoMode(checked)
        else:
            Internal.autoPlay = checked
            if checked and Internal.db:
                self.startPlayingGame()

    def updateGUI(self) ->None:
        """update some actions, all auxiliary windows and the statusbar"""
        if not isAlive(self):
            return
        self.setCaption('')
        for action in [self.actionScoreGame, self.actionPlayGame]:
            action.setEnabled(not bool(self.scene))
        self.actionAbortGame.setEnabled(bool(self.scene))
        scene = self.scene
        if isAlive(scene):
            assert scene
            scene.updateSceneGUI()

    @afterQueuedAnimations  # type:ignore[arg-type]
    def changeAngle(self, deferredResult:'Request', unusedButtons:Any=None, unusedModifiers:Any=None) ->None: # pylint: disable=unused-argument
        """change the lightSource"""
        if self.scene:
            with AnimationSpeed():
                self.scene.changeAngle()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only



Here we define replacement classes for the case that we have no
python interface to KDE.

"""


import os
import sys
from locale import setlocale, LC_ALL
from locale import _parse_localename  # type: ignore
from typing import Any, List, Generator, Optional

# pylint: disable=wrong-import-order

# here come the replacements:

# pylint: disable=wildcard-import,unused-wildcard-import
from qt import *

from common import Internal, Debug
from util import uniqueList

import gettext


try:
    from kdepaths import LOCALEPATH
except ImportError:
    LOCALEPATH = None

__all__ = ['i18n', 'i18nc', 'qi18nc', 'i18nE', 'i18ncE', 'KDETranslator', 'MLocale']


ENGLISHDICT = {}

def __insertArgs(translatedTemplate:str, *args:Any) ->str:
    """
    put format arguments into the translated template.
    KDE semantics markup is removed.

    @param translatedTemplate: The translated string template.
    @type translatedTemplate: C{str}
    @param args: The format arguments
    @type args: A list or tuple of C{str}

    @return: The formatted translated text.
    @rtype: C{str}
    """

    if '\004' in translatedTemplate:
        translatedTemplate = translatedTemplate.split('\004')[1]
    result = translatedTemplate
    if '%' in result:
        for idx in range(len(args)):
            result = result.replace(f'%{int(idx + 1)}', '{%d}' % idx)
        result = result.format(*args)
    for ignore in ['numid', 'filename', 'interface']:
        result = result.replace(f'<{ignore}>', '')
        result = result.replace(f'</{ignore}>', '')
    return result


def i18n(englishIn:str, *args:Any) ->str:
    """
    Translate. Since this is a 1:1 replacement for the
    corresponding KDE function, it accepts only C{str}.

    @param englishIn: The english template.
    @type englishIn: C{str}
    @return: The translated text, args included.
    @rtype: C{str}
    """
    if not Debug.neutral and englishIn:
        _ = MLocale.gettext(englishIn)
    else:
        _ = englishIn
    if not args:
        ENGLISHDICT[_] = englishIn
    result = __insertArgs(_, *args)
    return result


def i18nc(context:str, englishIn:str, *args:Any) ->str:
    """
    Translate. Since this is a 1:1 replacement for the
    corresponding KDE function, it accepts only C{str}.

    @param context: The context of this string.
    @type context: C{str}
    @param englishIn: The english template.
    @type englishIn: C{str}
    @return: The translated text, args included.
    @rtype: C{str}
    """
    # The \004 trick is taken from kdecore/localization/gettext.h,
    # definition of pgettext_aux"""
    if Debug.neutral:
        return __insertArgs(englishIn, *args)
    withContext = '\004'.join([context, englishIn])
    _ = MLocale.gettext(withContext)
    if _ == withContext:
        # try again without context
        _ = MLocale.gettext(englishIn)
    if not args:
        ENGLISHDICT[_] = englishIn
    return __insertArgs(_, *args)

def qi18nc(context:str, englishIn:str, *args:Any) ->str:
    """This uses the Qt translation files"""
    if Internal.app is None:
        _ = englishIn
    else:
        _ = Internal.app.translate(context.encode(), englishIn.encode())
    return __insertArgs(_, *args)

def i18nE(englishText:str) ->str:
    """use this if you want to get the english text right now but still have the string translated"""
    return englishText


def i18ncE(unusedContext:str, englishText:str) ->str:
    """use this if you want to get the english text right now but still have the string translated"""
    return englishText

def english(i18nstring:str) ->str:
    """translate back from local language"""
    return ENGLISHDICT.get(i18nstring, i18nstring)


class KDETranslator(QTranslator):

    """we also want Qt-only strings translated. Make Qt call this
    translator for its own strings. Use this with qi18nc()"""

    def __init__(self, parent:QObject) ->None:
        super().__init__(parent)

    def translate(self, context:str, text:str,  # type:ignore[override]
        disambiguation:Optional[bytes]=None, numerus:int=-1) ->str:
        """context should be the class name defined by Qt.
        PyQt uses str, Pyside uses bytes - so just ignore typing warnings"""
        # Qt doc says str but on Debian Bookworm, .pyi says bytes
        if Debug.neutral:
            return text  # type:ignore[return-value]
        result = super().translate(context, text, disambiguation, numerus)  # type:ignore[arg-type]
        if result:
            return result  # type:ignore[return-value]
        if not MLocale.currentLanguages():
            # when starting kajongg.py, Qt loads translators for the system default
            # language. I do not know how to avoid that. And I cannot delete
            # translators I do not own. So if we want no translation, just return text.
            # But we still need to install our own QTranslator overriding the ones
            # mentioned above. It will never find a translation, so we come here.
            assert Internal.app.translators == [self]
        return text  # type:ignore[return-value]



class MLocale:
    """xxxx"""

    __cached_availableLanguages = None

    translation:Optional[gettext.NullTranslations]  = None  # pylint:disable=used-before-assignment

    @classmethod
    def gettext(cls, txt:str) ->str:
        """with lazy installation of languages"""
        if cls.translation is None and Internal.kajonggrc:
            cls.installTranslations()
        if cls.translation is None:
            return txt
        return cls.translation.gettext(txt)

    @classmethod
    def installTranslations(cls) ->None:
        """install translations"""
        if sys.platform == 'win32':
            # on Linux, QCoreApplication initializes locale but not on Windows.
            # This is actually documented for QCoreApplication
            setlocale(LC_ALL, '')
        languages = cls.currentLanguages()
        cls.translation = gettext.NullTranslations()
        if Debug.i18n:
            Internal.logger.debug('Trying to install translations for %s', ','.join(languages))
        for language in languages:
            for context in ('kajongg', 'libkmahjongg5', 'kxmlgui5', 'kconfigwidgets5', 'libc'):
                directories = cls.localeDirectories()
                for resourceDir in directories:
                    try:
                        cls.translation.add_fallback(gettext.translation(
                            context, resourceDir, languages=[language]))
                        if Debug.i18n:
                            Internal.logger.debug(
                                'Found %s translation for %s in %s', language, context, resourceDir)
                        break
                    except IOError as _:
                        if Debug.i18n:
                            Internal.logger.debug(str(_))
        cls.translation.install()

    @staticmethod
    def localeDirectories() ->List[str]:
        """hard coded paths to i18n directories, all are searched"""
        candidates = (
            'share/locale', '/usr/local/share/locale', '/usr/share/locale',
            os.path.join(os.path.dirname(sys.argv[0]), 'share/locale'))
        result = [x for x in candidates if os.path.exists(x)]
        if not result and Debug.i18n:
            Internal.logger.debug('no locale path found. We have:%s', os.listdir('.'))

        if LOCALEPATH and os.path.exists(LOCALEPATH):
            result.insert(0, LOCALEPATH)
        return result

    @classmethod
    def currentLanguages(cls) ->List[str]:
        """the currently used languages, primary first"""
        try:
            languages = Internal.kajonggrc.group('Locale').readEntry('Language')
        except AttributeError:
            return []
        if not languages:
            return []
        languages = languages.split(':')
        if 'en_US' in languages:
            languages.remove('en_US')
        return languages

    @staticmethod
    def extendRegionLanguages(languages:List[str]) ->Generator[str, None, None]:
        """for de_DE, return de_DE, de"""
        for lang in languages:
            if lang is not None:
                yield lang
                if '_' in lang:
                    yield lang.split('_')[0]

    @staticmethod
    def get_localenames() ->List[str]:
        """parse environment variables"""
        result = []
        for variable in ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG'):
            try:
                localename = os.environ[variable]
                if localename is None:
                    raise ValueError(f'{variable} is None')
            except KeyError:
                continue
            else:
                if variable == 'LANGUAGE':
                    result.extend(localename.split(':'))
                else:
                    result.append(localename)
        if Debug.i18n:
            Internal.logger.debug('get_localenames: %s', format(','.join(result)))
        return result

    @classmethod
    def availableLanguages(cls) ->str:
        """see python lib, getdefaultlocale (which only returns the first one)"""
        if cls.__cached_availableLanguages:
            return cls.__cached_availableLanguages
        localenames = cls.get_localenames()
        languages = []
        for _ in localenames:
            if _ is None:
                continue
            _ = _parse_localename(_)
            if _[0]:
                languages.append(_[0])
        if Debug.i18n:
            Internal.logger.debug('languages from locale: %s', ','.join(languages) if languages else None)
            Internal.logger.debug('looking for translations in %s', ','.join(cls.localeDirectories()))
        installed_languages = set()
        for resourceDir in cls.localeDirectories():
            installed_languages |= set(os.listdir(resourceDir))
        for sysLanguage in sorted(installed_languages):
            if cls.__isLanguageInstalledForKajongg(sysLanguage):
                languages.append(sysLanguage)

        if languages:
            languages = uniqueList(cls.extendRegionLanguages(languages))
            languages = [x for x in languages if cls.isLanguageInstalled(x)]
        if 'en_US' not in languages:
            languages.extend(['en_US', 'en'])
        if Debug.i18n:
            Internal.logger.debug('languages available: %s', ':'.join(languages) if languages else None)
        cls.__cached_availableLanguages = ':'.join(languages)
        return cls.__cached_availableLanguages

    @classmethod
    def availableLanguages_(cls) ->str:
        """like availableLanguages but if xx_yy exists, exclude xx"""
        languages = set(cls.availableLanguages().split(':'))
        for _ in list(languages):
            if '_' in _ and _[:2] in languages:
                languages.remove(_[:2])
        return ':'.join(sorted(languages))

    @classmethod
    def isLanguageInstalled(cls, lang:str) ->bool:  # TODO: should be is Available
        """is any translation available for lang?"""
        if lang == 'en_US':
            return True
        for directory in cls.localeDirectories():
            if os.path.exists(os.path.join(directory, lang)):
                return True
        return False

    @classmethod
    def __isLanguageInstalledForKajongg(cls, lang:str) ->bool:
        """see kdelibs, KCatalog::catalogLocaleDir"""
        for directory in cls.localeDirectories():
            _ = os.path.join(directory, lang, 'LC_MESSAGES', 'kajongg.mo')
            if os.path.exists(_):
                if Debug.i18n:
                    Internal.logger.debug('language %s installed in %s', lang, _)
                return True
        return False



# -*- coding: utf-8 -*-
# pylint: skip-file

"""
Copyright (c) 2001-2011 Twisted Matrix Laboratories <twisted-python@twistedmatrix.com>
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

"""
This module provides support for Twisted to be driven by the Qt mainloop.

In order to use this support, simply do the following::
    |  app = QApplication(sys.argv) # your code to init Qt
    |  import qtreactor
    |  qtreactor.install()

alternatively:

    |  from twisted.application import reactors
    |  reactors.installReactor('qt')

Then use twisted.internet APIs as usual.  The other methods here are not
intended to be called directly.

If you don't instantiate a QApplication or QCoreApplication prior to
installing the reactor, a QCoreApplication will be constructed
by the reactor.  QCoreApplication does not require a GUI so trial testing
can occur normally.

Twisted can be initialized after QApplication.exec() with a call to
reactor.runReturn().  calling reactor.stop() will unhook twisted but
leave your Qt application running

API Stability: stable

Maintainer: U{Glenn H Tarbox, PhD<mailto:glenn@tarbox.org>}

Previous maintainer: U{Itamar Shtull-Trauring<mailto:twisted@itamarst.org>}
Original port to QT4: U{Gabe Rudy<mailto:rudy@goldenhelix.com>}
Subsequent port by therve
"""

import sys
from _collections_abc import dict_keys
from typing import List, Any, Optional, Callable, Dict, TYPE_CHECKING, Type, Tuple, Union

from zope.interface import implementer
from twisted.internet.interfaces import IReactorFDSet
from twisted.python import log, runtime
from twisted.internet import base, posixbase, unix

from qt import QSocketNotifier, QObject, QTimer, QCoreApplication
from qt import QEvent, QEventLoop


#if TYPE_CHECKING:

class TwistedSocketNotifier(QObject):

    """
    Connection between an fd event and reader/writer callbacks.
    """

    def __init__(self, parent:Optional[QObject], reactor:'QtReactor', watcher:unix.Client, socketType:QSocketNotifier.Type) ->None:
        super().__init__(parent)
        self.reactor = reactor
        self.watcher:Optional[unix.Client] = watcher
        fd = watcher.fileno()
        self.fn:Optional[Callable[...,None]]
        self.notifier = QSocketNotifier(fd, socketType, parent)
        self.notifier.setEnabled(True)
        if socketType == QSocketNotifier.Type.Read:
            self.fn = self.read
        else:
            self.fn = self.write
        self.notifier.activated.connect(self.fn)

    def shutdown(self) ->None:
        self.notifier.setEnabled(False)
        assert self.fn
        self.notifier.activated.disconnect(self.fn)
        self.fn = None
        self.watcher = None
        self.notifier.deleteLater()
        self.deleteLater()

    def read(self, fd:int) ->None:
        if not self.watcher:
            return
        w = self.watcher
        # doRead can cause self.shutdown to be called so keep a reference to
        # self.watcher

        def _read() ->None:
            # Don't call me again, until the data has been read
            self.notifier.setEnabled(False)
            why = None
            try:
                why = w.doRead()
                inRead = True
            except:
                inRead = False
                log.err()
                why = sys.exc_info()[1]
            if why:
                self.reactor._disconnectSelectable(w, why, inRead)
            elif self.watcher:
                self.notifier.setEnabled(
                    True)  # Re enable notification following successful read
            self.reactor._iterate(fromqt=True)
        log.callWithLogger(w, _read)

    def write(self, fd:int) ->None:
        if not self.watcher:
            return
        w = self.watcher

        def _write() ->None:
            why = None
            self.notifier.setEnabled(False)

            try:
                why = w.doWrite()
            except:
                log.err()
                why = sys.exc_info()[1]
            if why:
                self.reactor._disconnectSelectable(w, why, False)
            elif self.watcher:
                self.notifier.setEnabled(True)
            self.reactor._iterate(fromqt=True)
        log.callWithLogger(w, _write)


@implementer(IReactorFDSet)
class QtReactor(posixbase.PosixReactorBase):

    def __init__(self) ->None:
#        self._reads:Dict[posixbase._UnixWaker, TwistedSocketNotifier] = {}
        self._reads:Dict[unix.Client, TwistedSocketNotifier] = {}
        self._writes:Dict[Any, Any] = {}
        self._notifiers:Dict[Any, Any] = {}
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self.iterate)  # type:ignore[has-type]

        self.qApp:QCoreApplication
        if QCoreApplication.instance() is None:
            # Application Object has not been started yet
            self.qApp = QCoreApplication([])
            self._ownApp = True
        else:
            _ = QCoreApplication.instance()
            assert _
            self.qApp = _
            self._ownApp = False
        self._blockApp:Union[QCoreApplication, QEventLoop, None] = None
        posixbase.PosixReactorBase.__init__(self)

    def _add(self, xer:unix.Client, primary:Dict[unix.Client, TwistedSocketNotifier], typus:QSocketNotifier.Type) ->None:
        """
        Private method for adding a descriptor from the event loop.

        It takes care of adding it if  new or modifying it if already added
        for another state (read -> read/write for example).
        """
        if xer not in primary:
            primary[xer] = TwistedSocketNotifier(None, self, xer, typus)

    def addReader(self, reader:unix.Client) ->None:
        """
        Add a FileDescriptor for notification of data available to read.
        """
        self._add(reader, self._reads, QSocketNotifier.Type.Read)

    def addWriter(self, writer:unix.Client) ->None:
        """
        Add a FileDescriptor for notification of data available to write.
        """
        self._add(writer, self._writes, QSocketNotifier.Type.Write)

    def _remove(self, xer:unix.Client, primary:Dict[unix.Client, TwistedSocketNotifier]) ->None:
        """
        Private method for removing a descriptor from the event loop.

        It does the inverse job of _add, and also add a check in case of the fd
        has gone away.
        """
        if xer in primary:
            notifier = primary.pop(xer)
            notifier.shutdown()

    def removeReader(self, reader:unix.Client) ->None:
        """
        Remove a Selectable for notification of data available to read.
        """
        self._remove(reader, self._reads)

    def removeWriter(self, writer:unix.Client) ->None:
        """
        Remove a Selectable for notification of data available to write.
        """
        self._remove(writer, self._writes)

    def removeAll(self) ->List[Any]:
        """
        Remove all selectables, and return a list of them.
        """
        rv = self._removeAll(self._reads, self._writes)
        return rv

    def getReaders(self) ->Any:#dict_keys[Any, TwistedSocketNotifier]:
        result =self._reads.keys()
        print('result:',type(result))
        return result 

    def getWriters(self) ->Any:#dict_keys[Any, TwistedSocketNotifier]:
        return self._writes.keys()

    def callLater(self, delay:float, *args:Any, **kargs:Any) ->base.DelayedCall:
        rval = super(QtReactor, self).callLater(delay, *args, **kargs)
        self.reactorInvocation()
        return rval

    def reactorInvocation(self) ->None:
        self._timer.stop()
        self._timer.setInterval(0)
        self._timer.start()

    def _iterate(self, delay:Optional[int]=None, fromqt:bool=False) ->None:
        """See twisted.internet.interfaces.IReactorCore.iterate.
        """
        self.runUntilCurrent()
        self.doIteration(delay, fromqt)

    iterate = _iterate  # type:ignore[assignment]

    def doIteration(self, delay:Optional[float]=None, fromqt:bool=False) ->None:
        'This method is called by a Qt timer or by network activity on a file descriptor'

        if not self.running and self._blockApp:
            self._blockApp.quit()
        self._timer.stop()
        delay = max(int(delay or 0), 1)
        if not fromqt:
            self.qApp.processEvents(QEventLoop.ProcessEventsFlag.AllEvents, delay * 1000)
        timeout = self.timeout() or 0.1
        self._timer.setInterval(int(timeout * 1000))
        self._timer.start()

    def runReturn(self, installSignalHandlers:bool=True) ->None:
        self.startRunning(installSignalHandlers=installSignalHandlers)
        self.reactorInvocation()

    def run(self, installSignalHandlers:bool=True) ->None:
        if self._ownApp:
            self._blockApp = self.qApp
        else:
            self._blockApp = QEventLoop()
        self.runReturn()
        assert self._blockApp
        self._blockApp.exec()


class QtEventReactor(QtReactor):

    def __init__(self, *args:Any, **kwargs:Any) ->None:
        self._events:Dict[QEvent, Tuple[int, str]] = {}
        super(QtEventReactor, self).__init__()

    def addEvent(self, event:QEvent, fd:int, action:str) ->None:
        """
        Add a new win32 event to the event loop.
        """
        self._events[event] = (fd, action)

    def removeEvent(self, event:QEvent) ->None:
        """
        Remove an event.
        """
        if event in self._events:
            del self._events[event]

    def _runAction(self, action:str, fd:int) ->None:
        try:
            closed = getattr(fd, action)()
        except:
            closed = sys.exc_info()[1]
            log.deferr()

        if closed:
            self._disconnectSelectable(fd, closed, action == 'doRead')

    def timeout(self) ->Optional[float]:
        t = super(QtEventReactor, self).timeout()
        if t is not None:
            return min(t, 0.01)
        return None

    def iterate(self, delay:Optional[int]=None) ->None:  # type:ignore
        """See twisted.internet.interfaces.IReactorCore.iterate
        """
        self.runUntilCurrent()
        self.doIteration(delay)


def posixinstall() ->None:
    """
    Install the Qt reactor.
    """
    p = QtReactor()
    from twisted.internet.main import installReactor
    installReactor(p)


def win32install() ->None:
    """
    Install the Qt reactor.
    """
    p = QtEventReactor()
    from twisted.internet.main import installReactor
    installReactor(p)


if sys.platform == 'win32':
    install = win32install
else:
    install = posixinstall


__all__ = ["install"]



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import weakref
from typing import TYPE_CHECKING, Optional, List, Union, Type
from itertools import zip_longest
from twisted.internet.defer import Deferred

from common import ReprMixin
from tile import Piece, Tile, TileList

if TYPE_CHECKING:
    from game import Game
    from deferredutil import Request
    from uitile import UITile

class WallEmpty(Exception):

    """exception when trying to get a tile off the empty wall"""


class KongBox:

    """a non-ui kong box"""

    def __init__(self) ->None:
        self._tiles:List[Union[Piece, 'UITile']] = []

    def fill(self, tiles:List[Union[Piece, 'UITile']]) ->None:
        """fill the box"""
        self._tiles = tiles

    def pop(self, count:int) ->List[Union[Piece, 'UITile']]:
        """get count tiles from kong box"""
        if len(self._tiles) < count:
            raise WallEmpty
        tiles = self._tiles[-count:]
        self._tiles = self._tiles[:-count]
        return tiles

    def __getitem__(self, index:int) ->Union[Piece, 'UITile']:
        return self._tiles[index]

    def __len__(self) ->int:
        """# of tiles in kong box"""
        return len(self._tiles)


class Wall(ReprMixin):

    """represents the wall with four sides. self.wall[] indexes them
    counter clockwise, 0..3. 0 is bottom.
    Wall.tiles always holds references to all tiles in the game even
    when they are used"""
    tileClass : Type = Piece
    kongBoxClass = KongBox

    def __init__(self, game:'Game') ->None:
        """init and position the wall"""
        self._game = weakref.ref(game)  # avoid cycles for garbage collection
        self.tiles:List[Union[Piece, 'UITile']] = [self.tileClass(Piece.unknownStr)
                      for _ in range(game.fullWallSize)]
        self.living:List[Union[Piece, 'UITile']] = []
        self.kongBox = self.kongBoxClass()
        assert len(self.tiles) % 8 == 0

    @property
    def game(self) ->Optional['Game']:
        """hide the fact that this is a weakref"""
        return self._game()

    @staticmethod
    def __nameTile(tile:Union['UITile', Piece], element:'Tile') ->Union['UITile', Piece]:
        """define what tile this is"""
        if tile.__class__.__name__ == 'UITile':
            tile.change_name(element)
        else:
            tile = tile.change_name(element)  # type:ignore[assignment]
        return tile

    def deal(self, tiles:Union[TileList, Tile, None]=None, deadEnd:bool=False) ->List[Union['UITile', Piece]]:
        """deal tiles. May raise WallEmpty.
        Returns a list of tiles"""
        if isinstance(tiles, Tile):
            tiles = TileList(tiles)
        elif tiles is None:
            tiles = TileList(Tile.none)
        elif len(tiles) == 0:
            raise ValueError('Wall.deal() does not accept empty TileList()')
        count = len(tiles)
        if deadEnd:
            dealTiles = self.kongBox.pop(count)
            if len(self.kongBox) % 2 == 0:
                self._placeLooseTiles()
        else:
            if len(self.living) < count:
                raise WallEmpty
            dealTiles = self.living[:count]
            self.living = self.living[count:]
        return [self.__nameTile(x[0], x[1]) for x in zip_longest(dealTiles, tiles)]

    def build(self, shuffleFirst:bool=False) ->Deferred:  # pylint:disable=unused-argument
        """virtual: build visible wall"""
        return Deferred()

    def _placeLooseTiles(self, deferredResult:Optional['Request']=None) ->None:
        """to be done only for UIWall"""

    def decorate4(self, deferredResult:Optional['Request']=None) ->None:
        """virtual: show player info on the wall"""

    def hide(self) ->None:
        """virtual: hide all four walls and their decorators"""

    def divide(self) ->None:
        """divides a wall, building a living end and a dead end"""
        # neutralise the different directions of winds and removal of wall
        # tiles
        assert self.game
        assert self.game.divideAt is not None
        # shift tiles: tile[0] becomes living end
        self.tiles[:] = self.tiles[
            self.game.divideAt:] + \
            self.tiles[0:self.game.divideAt]
        kongBoxSize = self.game.ruleset.kongBoxSize  # type:ignore[attr-defined]
        self.living = self.tiles[:-kongBoxSize]
        boxTiles = self.tiles[-kongBoxSize:]
        for pair in range(kongBoxSize // 2):
            boxTiles = boxTiles[:pair * 2] + [
                boxTiles[pair * 2 + 1],
                boxTiles[pair * 2]] + \
                boxTiles[pair * 2 + 2:]
        self.kongBox.fill(boxTiles)

    def __len__(self) ->int:
        """Proxy for len(tiles)"""
        return len(self.tiles)

    @staticmethod
    def debug_name() ->str:
        """name for debug messages"""
        return '4sided wall'

    def __str__(self) ->str:
        """for debugging"""
        return self.debug_name()



# -*- coding: utf-8 -*-
"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import weakref
from typing import Optional, TYPE_CHECKING, Dict, Any, Union, List, Tuple

from common import ReprMixin, Internal
from message import Message
from wind import Wind
from tile import Tile, TileTuple, Meld, MeldList

if TYPE_CHECKING:
    from tilesource import TileSource
    from player import PlayingPlayer


class Move(ReprMixin):  # pylint: disable=too-many-instance-attributes
    """used for decoded move information from the game server"""

    def __init__(self, player:Optional['PlayingPlayer'],
        command:Union[Message, str], kwargs:Dict[Any,Any]) ->None:

        # pylint: disable=too-many-statements

        self.message:Message
        if isinstance(command, Message):
            self.message = command
        else:
            assert isinstance(command, str) # for mypy
            self.message = Message.defined[command]
        self.tile:Tile
        self.lastTile:Tile
        self.withDiscardTile:Tile
        self.tiles:TileTuple
        self.table = None
        self.notifying = False
        self._player = weakref.ref(player) if player else None
        self.token = kwargs['token']
        self.kwargs = kwargs.copy()
        del self.kwargs['token']
        self.score = None
        self.lastMeld:Meld
        self.meld:Meld
        self.melds:MeldList
        self.exposedMeld:Meld
        self.source:'TileSource'
        self.md5sum:str
        self.wantedGame:str
        self.show:bool
        self.shouldSave:bool
        self.playerNames:List[Tuple[Wind, str]]
        self.deadEnd:bool
        self.mustRotateWinds:bool
        self.divideAt:int
        self.tableid:int
        self.gameid:int
        for key, value in kwargs.items():
            assert not isinstance(value, bytes), f'value is bytes:{repr(value)}'
            if value is None:
                self.__setattr__(key, None)
            else:
                if key.lower().endswith('tile'):
                    self.__setattr__(key, Tile(value))
                elif key.lower().endswith('tiles'):
                    self.__setattr__(key, TileTuple(value))
                elif key.lower().endswith('meld'):
                    self.__setattr__(key, Meld(value))
                elif key.lower().endswith('melds'):
                    self.__setattr__(key, MeldList(value))
                elif key == 'playerNames':
                    if Internal.isServer:
                        self.__setattr__(key, value)
                    else:
                        self.__setattr__(key, self.__convertWinds(value))
                else:
                    self.__setattr__(key, value)

    @staticmethod
    def __convertWinds(tuples:List[Tuple[str, str]]) ->List[Tuple[Wind, str]]:
        """convert wind strings to Wind objects"""
        result = []
        for wind, name in tuples:
            _  = Wind(wind), name
            result.append(_)
        return result

    @property
    def player(self) ->Optional['PlayingPlayer']:
        """hide weakref"""
        return self._player() if self._player else None

    def __str__(self) ->str:
        return f'{self.player!r} {self.message!r} {self.kwargs!r}'



# -*- coding: utf-8 -*-

"""
Authors of original libkmahjongg in C++:
    Copyright (C) 1997 Mathias Mueller <in5y158@public.uni-hamburg.de>
    Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

this adapted python code:
    Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import os
import sys

from typing import Optional, Generator, List, Dict, cast

from qt import QStandardPaths
from log import logWarning, logException, logError
from kde import KConfig
from mi18n import i18n

RESOURCEFORMAT = 1 # as long as backgrounds and tilesets are synchronous in their versions


class Resource:

    """Common code for backgrounds and tilesets"""

    resourceName : Optional[str] = None # to be overridden in Tileset and Background
    configGroupName : str

    """represents a complete tileset"""

    cache : Optional[Dict[str, 'Resource']] = None  # common cache: tiles and background must not share identical names!

    def __new__(cls, name:Optional[str]=None) ->'Resource':
        if cls.cache is None:
            cls.cache = {}
            cls.loadAll()
        if name is None:
            return cls.available()[0]
        return cls.cache.get(name) or cls.cache.get(cls.__name(name)) or cls.__build(name)

    @classmethod
    def __directories(cls) ->Generator[str, None,None]:
        """where to look for resources"""
        result = QStandardPaths.locateAll(
            QStandardPaths.StandardLocation.GenericDataLocation,
            f'kmahjongglib/{cls.resourceName}s', QStandardPaths.LocateOption.LocateDirectory)
        if not result:
            logError('cannot find kmahjongg resources. Is kdegames-mahjongg-data installed?', showStack=False)
            sys.exit(1)
        result.insert(0, os.path.join('share', 'kmahjongglib', f'{cls.resourceName}s'))
        return (x for x in result if os.path.exists(x))

    @classmethod
    def locate(cls, which:str) ->Optional[str]:
        """locate the file with a resource"""
        for directory in cls.__directories():
            path = os.path.join(directory, which)
            if os.path.exists(path):
                return path
        logException(f'cannot find kmahjongg{cls.resourceName} {which} in {cls.__directories()}')
        return None

    @classmethod
    def loadAll(cls) ->None:
        """loads all available resources into cache"""
        resourceDirectories = cls.__directories()
        for directory in resourceDirectories:
            for name in os.listdir(directory):
                if name.endswith('.desktop'):
                    if not name.endswith('alphabet.desktop') and not name.endswith('egypt.desktop'):
                        cls(os.path.join(directory, name))

    @classmethod
    def available(cls) ->List['Resource']:
        """ready for the selector dialog, default first"""
        cls.loadAll()
        assert cls.cache is not None
        return sorted(set(cls.cache.values()), key=lambda x: x.desktopFileName != 'default')

    @classmethod
    def __noTilesetFound(cls) ->None:
        """No resources found"""
        directories = '\n\n' + '\n'.join(cls.__directories())
        logException(
            i18n(
                'cannot find any %1 in the following directories, '
                'is libkmahjongg installed?', cls.resourceName) + directories) # TODO: nicht schoen

    @staticmethod
    def __name(path:str) ->str:
        """extract the name from path: this is the filename minus the .desktop ending"""
        return os.path.split(path)[1].replace('.desktop', '')

    @classmethod
    def __build(cls, name:str) ->'Resource':
        """build a new Resource. name is either a full file path or a desktop name. None stands for 'default'."""
        result = object.__new__(cls)
        if os.path.exists(name):
            result.path = name
            result.desktopFileName = cls.__name(name)
        else:
            result.desktopFileName = name or 'default'
            path = cls.locate(result.desktopFileName + '.desktop')
            if not path:
                path = cls.locate('default.desktop')
                result.desktopFileName = 'default'
                if not path:
                    cls.__noTilesetFound()
                else:
                    logWarning(i18n('cannot find %1, using default', name))
            assert path  # for mypy
            result.path = path

        assert cls.cache is not None
        cls.cache[result.desktopFileName] = result  # pylint:disable=unsupported-assignment-operation
        cls.cache[result.path] = result  # pylint:disable=unsupported-assignment-operation
        return result

    def __init__(self, unusedName:Optional[str]=None) ->None:
        """continue __build"""
        self.path: str
        self.desktopFileName: str
        self.group = KConfig(self.path).group(self.configGroupName)

        self.name = cast(str, self.group.readEntry("Name")) or i18n("unknown name")
        self.author = cast(str, self.group.readEntry("Author")) or i18n("unknown author")
        self.description = cast(str, self.group.readEntry(
            "Description")) or i18n(
                "no description available")
        self.authorEmail = cast(str, self.group.readEntry(
            "AuthorEmail")) or i18n(
                "no E-Mail address available")

        # Version control
        resourceVersion = self.group.readInteger("VersionFormat")
        # Format is increased when we have incompatible changes, meaning that
        # older clients are not able to use the remaining information safely
        if resourceVersion > RESOURCEFORMAT:
            logException(f'version file / program: {int(resourceVersion)}/{int(RESOURCEFORMAT)}')

    def __str__(self) ->str:
        return (f"{self.resourceName} id={int(id(self))} name={self.desktopFileName}, "
                f"name id={int(id(self.desktopFileName))}")

    @staticmethod
    def current() ->Optional['Resource']:
        """the currently wanted tileset. If not yet defined, do so"""
        return None



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import datetime

from typing import TYPE_CHECKING, Optional, List, Any, cast, Union

from qt import Qt, QAbstractTableModel, QModelIndex
from qt import QDialog, QDialogButtonBox, QWidget
from qt import QHBoxLayout, QVBoxLayout, QAbstractItemView
from qt import QItemSelectionModel, QGridLayout, QColor, QPalette

from mi18n import i18n, i18nc, i18nE
from kde import KApplication, KIcon, KDialogButtonBox

from genericdelegates import RichTextColumnDelegate

from log import logDebug
from statesaver import StateSaver
from rule import Ruleset
from guiutil import ListComboBox, MJTableView, decorateWindow
from differ import RulesetDiffer
from common import Internal, Debug
from modeltest import ModelTest
from chat import ChatMessage, ChatWindow

if TYPE_CHECKING:
    from client import ClientTable
    from qt import QObject, QEvent, QItemSelection, QPersistentModelIndex
    from qt import QPushButton
    from login import Url
    from humanclient import HumanClient
    from scene import PlayingScene

class TablesModel(QAbstractTableModel):

    """a model for our tables"""

    def __init__(self, tables:List['ClientTable'], parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        self.tables = tables
        assert isinstance(tables, list)

    def headerData(
            self, section:int,
            orientation:Qt.Orientation, role:int=Qt.ItemDataRole.DisplayRole) ->Any:
        """show header"""
        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                if section in [3, 4]:
                    return int(Qt.AlignmentFlag.AlignLeft)
                return int(Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter)
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if orientation != Qt.Orientation.Horizontal:
            return int(section + 1)
        result = ''
        if section < 5:
            result = [i18n('Table'),
                      '',
                      i18n('Players'),
                      i18nc('table status',
                            'Status'),
                      i18n('Ruleset')][section]
        return result

    def rowCount(self, parent:Union[QModelIndex,'QPersistentModelIndex']=QModelIndex()) ->int:
        """how many tables are in the model?"""
        if parent.isValid():
            # we have only top level items
            return 0
        return len(self.tables)

    def columnCount(self, unusedParent:Union[QModelIndex,'QPersistentModelIndex']=QModelIndex()) ->int:
        """for now we only have id (invisible), id (visible), players,
        status, ruleset.name.
        id(invisible) always holds the real id, also 1000 for suspended tables.
        id(visible) is what should be displayed."""
        return 5

    def data(self, index:Union[QModelIndex,'QPersistentModelIndex'], role:int=Qt.ItemDataRole.DisplayRole) ->Any:
        """score table"""
        # pylint: disable=too-many-branches,too-many-locals
        result:Any = None
        if role == Qt.ItemDataRole.TextAlignmentRole:
            if index.column() == 0:
                result = int(Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter)
            else:
                result = int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        if index.isValid() and (0 <= index.row() < len(self.tables)):
            table = self.tables[index.row()]
            if role == Qt.ItemDataRole.DisplayRole and index.column() in (0, 1):
                result = table.tableid
            elif role == Qt.ItemDataRole.DisplayRole and index.column() == 2:
                players = []
                zipped = list(zip(table.playerNames, table.playersOnline))
                for idx, pair in enumerate(zipped):
                    name, online = pair[0], pair[1]
                    if idx < len(zipped) - 1:
                        name += ', '
                    palette = KApplication.palette()
                    if online:
                        color = palette.color(
                            QPalette.ColorGroup.Active,
                            QPalette.ColorRole.WindowText).name()
                        style = f'font-weight:normal;font-style:normal;color:{color}'
                    else:
                        color = palette.color(
                            QPalette.ColorGroup.Disabled,
                            QPalette.ColorRole.WindowText).name()
                        style = f'font-weight:100;font-style:italic;color:{color}'
                    players.append(
                        f'<nobr style="{style}">' +
                        name +
                        '</nobr>')
                names = ''.join(players)
                result = names
            elif role == Qt.ItemDataRole.DisplayRole and index.column() == 3:
                status = table.status()
                if table.suspendedAt:
                    dateVal = ' ' + datetime.datetime.strptime(
                        table.suspendedAt,
                        '%Y-%m-%dT%H:%M:%S').strftime('%c')
                    status = 'Suspended'
                else:
                    dateVal = ''
                result = i18nc('table status', status) + dateVal
            elif index.column() == 4:
                if role == Qt.ItemDataRole.DisplayRole:
                    result = i18n((table.myRuleset if table.myRuleset else table.ruleset).name)
                elif role == Qt.ItemDataRole.ForegroundRole:
                    palette = KApplication.palette()
                    result = palette.windowText().color() if table.myRuleset else QColor('red')
        return result


class SelectRuleset(QDialog):

    """a dialog for selecting a ruleset"""

    def __init__(self, server:Optional['Url']=None) ->None:
        super().__init__(None)
        decorateWindow(self, i18n('Select a ruleset'))
        self.buttonBox = KDialogButtonBox(self)
        self.buttonBox.setStandardButtons(
            QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.cbRuleset = ListComboBox(Ruleset.selectableRulesets(server))
        self.grid = QGridLayout()  # our child SelectPlayers needs this
        self.grid.setColumnStretch(0, 1)
        self.grid.setColumnStretch(1, 6)
        vbox = QVBoxLayout(self)
        vbox.addLayout(self.grid)
        vbox.addWidget(self.cbRuleset)
        vbox.addWidget(self.buttonBox)


class TableList(QWidget):

    """a widget for viewing, joining, leaving tables"""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, client:'HumanClient'):
        super().__init__(None)
        self.autoStarted = False
        self.client = client
        self.setObjectName('TableList')
        self.resize(700, 400)
        self.view = MJTableView(self)
        self.__differ:RulesetDiffer
        self.debugModelTest:ModelTest
        self.requestedNewTable = False
        self.view.setItemDelegateForColumn(
            2,
            RichTextColumnDelegate(self.view))
        self.buttonBox = QDialogButtonBox(self)

        self.newButton = self.add_button(
            i18nc('allocate a new table', "&New"),
            i18n("Allocate a new table"),
            "document-new",
            QDialogButtonBox.ButtonRole.ActionRole,
            self.client.newTable)

        self.joinButton = self.add_button(
            i18n("&Join"),
            i18n("Join a table"),
            "list-add-user",
            QDialogButtonBox.ButtonRole.AcceptRole,
            client.joinTable)

        self.leaveButton = self.add_button(
            i18n("&Leave"),
            i18n("Leave a table"),
            "list-remove-user",
            QDialogButtonBox.ButtonRole.AcceptRole,
            self.leaveTable)

        self.compareButton = self.add_button(
            i18nc('Kajongg-Ruleset', 'Compare'),
            i18n('Compare the rules of this table with my own rulesets'),
            "preferences-plugin-script",
            QDialogButtonBox.ButtonRole.AcceptRole,
            self.compareRuleset)

        self.chatButton = self.add_button(
            i18n('&Chat'),
            "",
            "call-start",
            QDialogButtonBox.ButtonRole.AcceptRole,
            self.chat)

        self.startButton = self.add_button(
            i18n('&Start'),
            i18n("Start playing on a table. Empty seats will be taken by robot players."),
            "arrow-right",
            QDialogButtonBox.ButtonRole.AcceptRole,
            self.startGame)

        cmdLayout = QHBoxLayout()
        cmdLayout.addWidget(self.buttonBox)

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.addLayout(cmdLayout)
        self.setLayout(layout)

        self.view.doubleClicked.connect(client.joinTable)
        header = self.view.horizontalHeader()
        if header:
            StateSaver(self, header)
        self.__updateButtonsForNoTable()

    def add_button(self, text:str, tooltip:str, icon_name:str,
        role:QDialogButtonBox.ButtonRole, connector:Any) ->'QPushButton':
        """the wrapped add_button might return None, mypy does not like that"""
        result = self.buttonBox.addButton(text, role)
        assert result
        result.setIcon(KIcon(icon_name))
        result.setToolTip(tooltip)
        result.clicked.connect(connector)
        return result

    def hideEvent(self, unusedEvent:Optional['QEvent']) ->None:
        """table window hides"""
        scene = Internal.scene
        if scene:
            cast('PlayingScene', scene).startingGame = False
        model = cast(TablesModel, self.view.model())
        if model:
            for table in model.tables:
                if table.chatWindow:
                    table.chatWindow.hide()
                    table.chatWindow = None
        if scene:
            if not scene.game or scene.game.client != self.client:
                # do we still need this connection?
                self.client.logout()

    def chat(self) ->None:
        """chat. Only generate ChatWindow after the
        message has successfully been sent to the server.
        Because the server might have gone away."""
        msg:ChatMessage
        def initChat(_:Any) ->None:
            """now that we were able to send the message to the server
            instantiate the chat window"""
            assert table
            table.chatWindow = ChatWindow(table)
            table.chatWindow.receiveLine(msg)
        table = self.selectedTable()
        assert table
        if not table.chatWindow:
            line = i18nE('opens a chat window')
            msg = ChatMessage(
                table.tableid,
                table.client.name,
                line,
                isStatusMessage=True)
            cast('HumanClient', table.client).sendChat(
                msg).addCallback(
                    initChat).addErrback(
                        self.client.tableError)
        elif table.chatWindow.isVisible():
            table.chatWindow.hide()
        else:
            table.chatWindow.show()

    def show(self) ->None:
        """prepare the view and show it"""
        if self.client.hasLocalServer():
            title = i18n(
                'Local Games with Ruleset %1',
                self.client.ruleset.name)
        else:
            assert self.client.connection
            title = i18n('Tables at %1', self.client.connection.url)
        assert self.client.name
        decorateWindow(self, ' - '.join([self.client.name, title]))
        self.view.hideColumn(1)
        model = self.view.model()
        tableCount = model.rowCount() if model else 0
        self.view.showColumn(0)
        self.view.showColumn(2)
        self.view.showColumn(4)
        if tableCount or not self.client.hasLocalServer():
            super().show()
            if self.client.hasLocalServer():
                self.view.hideColumn(0)
                self.view.hideColumn(2)
                self.view.hideColumn(4)

    def selectTable(self, idx:int) ->None:
        """select table by idx"""
        self.view.selectRow(idx)
        self.updateButtonsForTable(self.selectedTable())

    def __updateButtonsForNoTable(self) ->None:
        """update button status when no table is selected"""
        self.joinButton.setEnabled(False)
        self.leaveButton.setVisible(False)
        self.compareButton.setVisible(False)
        self.startButton.setVisible(False)
        self.newButton.setToolTip(i18n("Allocate a new table"))
        self.joinButton.setText(i18n('&Join'))
        self.joinButton.setToolTip(i18n("Join a table"))
        self.startButton.setEnabled(False)
        self.compareButton.setEnabled(False)
        self.chatButton.setVisible(False)

    def updateButtonsForTable(self, table:Optional['ClientTable']=None) ->None:
        """update button status for the currently selected table"""
        if table is None:
            self.__updateButtonsForNoTable()
            return
        assert table
        suspended = bool(table.suspendedAt)
        running = table.running
        suspendedLocalGame = (
            suspended and table.gameid
            and self.client.hasLocalServer())
        self.joinButton.setEnabled(
            not running and
            not table.isOnline(self.client.name) and
            (self.client.name in table.playerNames) == suspended)
        self.leaveButton.setVisible(not suspendedLocalGame)
        self.compareButton.setVisible(not suspendedLocalGame)
        self.startButton.setVisible(not suspended)
        if suspendedLocalGame:
            self.newButton.setToolTip(i18n("Start a new game"))
            self.joinButton.setText(
                i18nc('resuming a local suspended game', '&Resume'))
            self.joinButton.setToolTip(
                i18n("Resume the selected suspended game"))
        else:
            self.newButton.setToolTip(i18n("Allocate a new table"))
            self.joinButton.setText(i18n('&Join'))
            self.joinButton.setToolTip(i18n("Join a table"))
        self.leaveButton.setEnabled(
            not running and not self.joinButton.isEnabled())
        self.startButton.setEnabled(
            not running and not suspendedLocalGame
            and self.client.name == table.playerNames[0])
        self.compareButton.setEnabled(table.myRuleset is None)
        self.chatButton.setVisible(not self.client.hasLocalServer())
        self.chatButton.setEnabled(
            not running
            and self.client.name in table.playerNames
            and sum(x.startswith('Robot ') for x in table.playerNames) < 3)
        if self.chatButton.isEnabled():
            self.chatButton.setToolTip(i18n(
                "Chat with others on this table"))
        else:
            self.chatButton.setToolTip(i18n(
                "For chatting with others on this table, "
                "please first take a seat"))

    def selectionChanged(self, selected:'QItemSelection', unusedDeselected:'QItemSelection') ->None:
        """update button states according to selection"""
        if selected.indexes():
            self.selectTable(selected.indexes()[0].row())

    def selectedTable(self) ->Optional['ClientTable']:
        """return the selected table"""
        sel_model = self.view.selectionModel()
        if sel_model:
            index = sel_model.currentIndex()
            model = cast(TablesModel, self.view.model())
            if index.isValid() and model:
                return model.tables[index.row()]
        return None

    def compareRuleset(self) ->None:
        """compare the ruleset of this table against ours"""
        table = self.selectedTable()
        if table:
            self.__differ = RulesetDiffer([table.ruleset], Ruleset.availableRulesets())
            self.__differ.show()

    def startGame(self) ->None:
        """start playing at the selected table"""
        table = self.selectedTable()
        if table:
            self.startButton.setEnabled(False)
            self.client.callServer(
                'startGame',
                table.tableid).addErrback(
                    self.client.tableError)

    def leaveTable(self) ->None:
        """leave a table"""
        table = self.selectedTable()
        if table:
            self.client.callServer(
                'leaveTable',
                table.tableid).addErrback(
                    self.client.tableError)

    def __keepChatWindows(self, tables:List['ClientTable']) ->None:
        """copy chatWindows from the old table list which will be
        thrown away"""
        model = cast(TablesModel, self.view.model())
        if model:
            chatWindows = {x.tableid: x.chatWindow for x in model.tables}
            unusedWindows = {x.chatWindow for x in model.tables}
            for table in tables:
                table.chatWindow = chatWindows.get(table.tableid, None)
                unusedWindows -= {table.chatWindow}
            for unusedWindow in unusedWindows:
                if unusedWindow:
                    unusedWindow.hide()

    def __preselectTableId(self, tables:List['ClientTable']) ->int:
        """which table should be preselected?
        If we just requested a new table:
          select first new table.
          Only in the rare case that two clients request a new table at
          the same moment, this could put the focus on the wrong table.
          Ignore that for now.
        else if we had one selected:
          select that again
        else:
          select first table"""
        if self.requestedNewTable:
            self.requestedNewTable = False
            model = cast(TablesModel, self.view.model())
            if model:
                oldIds = {x.tableid for x in model.tables}
                newIds = sorted({x.tableid for x in tables} - oldIds)
                if newIds:
                    return newIds[0]
        _ = self.selectedTable()
        if _:
            return _.tableid
        return 0

    def loadTables(self, tables:List['ClientTable']) ->None:
        """build and use a model around the tables.
        Show all new tables (no gameid given yet) and all suspended
        tables that also exist locally. In theory all suspended games should
        exist locally but there might have been bugs or somebody might
        have removed the local database like when reinstalling linux"""
        if not Internal.scene:
            return
        if Debug.table:
            for table in tables:
                if table.gameid and not table.gameExistsLocally():
                    logDebug(f'Table {table} does not exist locally')
        tables = [x for x in tables if not x.gameid or x.gameExistsLocally()]
        tables.sort(key=lambda x: x.tableid)
        preselectTableId = self.__preselectTableId(tables)
        self.__keepChatWindows(tables)
        model = TablesModel(tables)
        self.view.setModel(model)
        if Debug.modelTest:
            self.debugModelTest = ModelTest(model, self.view)
        selection = QItemSelectionModel(model, self.view)
        self.view.initView()
        self.view.setSelectionModel(selection)
        self.view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        selection.selectionChanged.connect(self.selectionChanged)
        if len(tables) == 1:
            self.selectTable(0)
            self.startButton.setFocus()
        elif not tables:
            self.newButton.setFocus()
        else:
            _ = [x for x in tables if x.tableid >= preselectTableId]
            self.selectTable(tables.index(_[0]) if _ else 0)
        self.updateButtonsForTable(self.selectedTable())
        self.view.setFocus()



# -*- coding: utf-8 -*-

"""
Copyright 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import weakref
from collections import defaultdict
from typing import Optional, TYPE_CHECKING, List, Dict, Union, cast

from qt import QColor
from tile import Tile, TileList, Meld, MeldList
from tileset import Tileset
from uitile import UITile
from hand import Hand
from board import Board
from sound import Sound

from log import logDebug
from common import Internal, Debug, isAlive, ReprMixin

if TYPE_CHECKING:
    from qt import QGraphicsSceneDragDropEvent
    from visible import VisiblePlayer
    from scene import PlayingScene
    from player import PlayingPlayer


class TileAttr(ReprMixin):

    """a helper class for syncing the hand board, holding relevant
    tile attributes.
    xoffset is expressed in number of tiles but may be
    fractional for adding distances between melds"""

    def __init__(self, hand:Union['HandBoard', UITile], meld:Optional[Meld]=None, idx:Optional[int]=None,
        xoffset:Optional[float]=None, yoffset:Optional[int]=None) ->None:
        if isinstance(hand, UITile):
# TODO only used for Bonus tiles, should split or eliminate this
            self.tile = hand.tile
            self.xoffset = hand.xoffset
            self.yoffset = hand.yoffset
            self.dark = hand.dark
            self.focusable = hand.focusable
        else:
            assert meld
            assert idx is not None
            self.tile = Tile(meld[idx])
            assert xoffset is not None # FIXME: overload __init__
            assert yoffset is not None
            self.xoffset = xoffset  # type:ignore[assignment]
            self.yoffset = yoffset
            self.dark = self._setDark()
            # dark and focusable are different in a ScoringHandBoard
            self.focusable = self._setFocusable(hand, meld, idx)
            if self.tile.name2() in Debug.focusable:
                logDebug(f'TileAttr {self.tile}:{self.focusable}')

    def _setDark(self) ->bool:
        """should the tile appear darker?"""
        return self.tile.isConcealed if self.yoffset == 0 else not self.tile.isKnown

    def _setFocusable(self, hand:'HandBoard', meld:Meld, idx:Optional[int]) ->bool: # pylint: disable=unused-argument
        """is it focusable?"""
        player = hand.player
        assert player
        assert player.game
        return (
            not self.tile.isBonus
            and self.tile.isKnown
            and player == player.game.activePlayer
            and player == player.game.myself
            and meld.isConcealed and not meld.isKong)

    def apply_to(self, board:Board, uiTile:UITile) -> None:
        """Change uiTile according to self"""
        uiTile.level = 0  # for tiles coming from the wall
        uiTile.change_name(self.tile)
        uiTile.setBoard(board, self.xoffset, self.yoffset)
        uiTile.dark = self.dark
        uiTile.focusable = self.focusable

    def __str__(self) ->str:
        assert self.xoffset is not None
        return (
            f"{self.tile} {self.xoffset:.2f}/{int(self.yoffset)}{' dark' if self.dark else ''}"
            f"{' focusable' if self.focusable else ''}")


class HandBoard(Board):

    """a board showing the tiles a player holds"""

    tileAttrClass = TileAttr
    penColor = QColor('blue')
    showShadowsBetweenRows = True

    def __init__(self, player:'VisiblePlayer') ->None:
        assert player
        self._player = weakref.ref(player)
        self.exposedMeldDistance:float = 0.15
        self.concealedMeldDistance:float = 0.0
        super().__init__(15.6, 2.0, Tileset.current())
        self.isHandBoard = True
        self.tileDragEnabled = False
        self.setParentItem(player.front)
        self.setPosition()
        self.setAcceptDrops(True)
        assert Internal.Preferences
        Internal.Preferences.addWatch(
            'rearrangeMelds', self.rearrangeMeldsChanged)
        self.rearrangeMeldsChanged(False, bool(Internal.Preferences.rearrangeMelds))
        Internal.Preferences.addWatch(
            'showShadows', self.showShadowsChanged)

    def computeRect(self) ->None:
        """also adjust the scale for maximum usage of space"""
        super().computeRect()
        rect = self.rect()
        rect.setWidth(rect.width() + 2 * self.tileset.shadowHeight())
        self.setRect(rect)

        if self.player:
            sideRect = self.player.front.boundingRect()
            boardRect = self.boundingRect()
            scale = ((sideRect.width() + sideRect.height())
                     / (boardRect.width() - boardRect.height()))
            self.setScale(scale)

    @property
    def player(self) ->Optional['VisiblePlayer']:
        """player is readonly and never None"""
        return self._player() if self._player else None

    # this is ordered such that pylint does not complain about
    # identical code in board.py

    def debug_name(self) ->str:
        """for debugging messages"""
        if self.player is None:
            return 'None'
        return self.player.name

    def showShadowsChanged(self, unusedOldValue:bool, newValue:bool) ->None:
        """Add or remove the shadows."""
        self.setPosition()

    def setPosition(self) ->None:
        """Position myself"""
        assert Internal.Preferences
        show = bool(Internal.Preferences.showShadows)
        if show:
            self.setTilePos(yHeight=1.5)
        else:
            self.setTilePos(yHeight=1.0)
        self.setBoardRect(15.6, 2.2 if show else 2.0)
        self._reload(self.tileset, showShadows=show)
        self.sync()

    def rearrangeMeldsChanged(self, unusedOldValue:bool, newValue:bool) ->None:
        """when True, concealed melds are grouped"""
        self.concealedMeldDistance = (
            self.exposedMeldDistance if newValue else 0.0)
        self._reload(self.tileset, self._lightSource)
        self.sync()

    def focusRectWidth(self) ->int:
        """how many tiles are in focus rect? We want to focus
        the entire meld"""
        # playing game: always make only single tiles selectable
        return 1

    def __str__(self) ->str:
        assert self.player
        return self.player.scoringString()

    def lowerHalfTiles(self) ->List[UITile]:
        """return a list with all single tiles of the lower half melds
        without boni"""
        return [x for x in self if x.yoffset > 0 and not x.isBonus]

    def newLowerMelds(self) ->MeldList:
        """a list of melds for the hand as it should look after sync"""
        return MeldList()

    def listNewTilePositions(self) ->List[TileAttr]:
        """return list(TileAttr) for all tiles except bonus tiles.
        The tiles are not associated to any board."""
        result:List[TileAttr] = []
        assert self.player
        newUpperMelds = list(self.player.exposedMelds)
        newLowerMelds = self.newLowerMelds()
        for yPos, melds in ((0, newUpperMelds), (1, newLowerMelds)):
            meldDistance = (self.concealedMeldDistance if yPos
                            else self.exposedMeldDistance)
            meldX:float = 0.0
            for meld in melds:
                for idx in range(len(meld)):
                    result.append(self.tileAttrClass(self, meld, idx, meldX, yPos))
                    meldX += 1
                meldX += meldDistance
        return sorted(result, key=lambda x: x.yoffset * 100 + x.xoffset)

    def __findMaxX(self, positions:List[TileAttr], row:int) ->float:
        """find the rightmost position for row"""
        x_values = [position.xoffset for position in positions if position.yoffset == row]
        return max(x_values) if x_values else 0.0

    def __placeBoniInRow(self, bonusTiles:List[UITile], after:List[float],
        bonusY:int, keepTogether:bool, maxTilesInRow:int) ->List[TileAttr]:
        """Try to place bonusTiles in upper or in lower row.
        tilePositions are the normal tiles, already placed.
        Placed boni are removed from bonusTiles.
        Returns all tiles that could be placed."""

        result:List[TileAttr] = []
        placeBoni = bonusTiles[:]
        while maxTilesInRow - len(placeBoni) < after[bonusY] + 1 + self.exposedMeldDistance:
            if keepTogether:
                return []
            placeBoni = placeBoni[:-1]
            if not placeBoni:
                return result
        for _ in placeBoni:
            bonusTiles.remove(_)
        xPos = maxTilesInRow - len(placeBoni)
        newBonusTiles = [self.tileAttrClass(x) for x in placeBoni]
        for bonus in sorted(newBonusTiles, key=lambda x: x.tile.key):
            bonus.xoffset, bonus.yoffset = xPos, bonusY
            bonus.dark = False
            result.append(bonus)
            after[bonusY] += 1
            xPos += 1
        return result

    def __newBonusPositions(self, bonusTiles:List[UITile], after:List[float]) ->List[TileAttr]:
        """return list(TileAttr)
        calculate places for bonus tiles. Try to put them all in one row,
        right adjusted. If necessary, extend to the right even
        outside of our board"""
        result:List[TileAttr] = []
        bonusTiles = bonusTiles[:]  # do not change passed list
        assert self.player
        assert self.player.game
        maxTilesInRow = self.player.game.ruleset.dealtTiles
        result.extend(self.__placeBoniInRow(bonusTiles, after, 0, True, maxTilesInRow))
        result.extend(self.__placeBoniInRow(bonusTiles, after, 1, True, maxTilesInRow))
        if len(bonusTiles):
            # we cannot place all bonus tiles in the same row!
            result.extend(self.__placeBoniInRow(bonusTiles, after, 0, False, maxTilesInRow))
            result.extend(self.__placeBoniInRow(bonusTiles, after, 1, False, maxTilesInRow))
        while len(bonusTiles):
            maxTilesInRow += 1
            assert maxTilesInRow < 99, f'cannot place {bonusTiles}'
            result.extend(self.__placeBoniInRow(bonusTiles, after, 0, False, maxTilesInRow))
            result.extend(self.__placeBoniInRow(bonusTiles, after, 1, False, maxTilesInRow))
        return result

    def placeTiles(self, tiles:List[UITile]) ->None:
        """tiles are all tiles for this board."""
        oldTiles = defaultdict(list)
        for uiTile in filter(lambda x: not x.isBonus, tiles):
            oldTiles[uiTile.tile].append(uiTile)
        matches:Dict[UITile, TileAttr] = {}
        newPositions = self.listNewTilePositions()
        for newPosition in newPositions:
            assert isinstance(newPosition.tile, Tile)
            candidates = oldTiles.get(newPosition.tile) \
                or oldTiles.get(newPosition.tile.swapped) \
                or oldTiles.get(Tile.unknown)
            if not candidates and not newPosition.tile.isKnown and oldTiles:
                # 13 orphans, robbing Kong, lastTile is single:
                # no oldTiles exist
                candidates = list(oldTiles.values())[0]
            if candidates:
                # no candidates happen when we move a uiTile within a board,
                # here we simply ignore existing tiles with no candidates
                candidates = sorted(
                    candidates, key=lambda x:
                    + abs(newPosition.yoffset - x.yoffset) * 100 # pylint: disable=cell-var-from-loop
                    + x.xoffset)
                # pylint is too cautious here. Check with later versions.
                match = candidates[0]
                matches[match] = newPosition
                oldTiles[match.tile].remove(match)
                if not oldTiles[match.tile]:
                    del oldTiles[match.tile]
        for uiTile, newPos in matches.items():
            newPos.apply_to(self, uiTile)
        after = list(self.__findMaxX(list(matches.values()), x) for x in (0, 1))
        self._placeBonusTiles(after, tiles)

    def _placeBonusTiles(self, after:List[float], tiles:List[UITile]) ->None:
        """Temporary code, directly after extraction from placeTiles()"""
        boni = list(sorted(filter(lambda x: x.isBonus, tiles), key=lambda x: hash(x.tile)))
        positions = self.__newBonusPositions(boni, after)
        for tile, position in zip(boni, positions):
            position.apply_to(self, tile)

    def sync(self, adding:Optional[List[UITile]]=None) ->None:
        """place all tiles in HandBoard.
        adding tiles: their board is where they come from. Those tiles
        are already in the Player tile lists.
        The sender board must not be self, see ScoringPlayer.moveMeld"""
        if self.empty and not adding:
            return
        allTiles = self[:]
        if adding:
            allTiles.extend(adding)
        self.placeTiles(allTiles)
        self.checkTiles()

    def checkTiles(self) ->None:
        """does the logical state match the displayed tiles?"""
        # FIXME: when exactly should I call this? afterQueuedAnimations does not help
        #    test case: scoring game. move meld from exposed to concealed using the mouse
        logExposed:TileList = TileList()
        physExposed:TileList = TileList()
        logConcealed:TileList = TileList()
        physConcealed:TileList = TileList()
        assert self.player
        for tile in self.player.bonusTiles:
            logExposed.append(tile)
        for uiTile in self:
            if uiTile.yoffset == 0 or uiTile.isBonus:
                physExposed.append(uiTile.tile)
            else:
                physConcealed.append(uiTile.tile)
        for meld in self.player.exposedMelds:
            logExposed.extend(meld)
        if self.player.concealedMelds:
            for meld in self.player.concealedMelds:
                logConcealed.extend(meld)
        else:
            logConcealed = TileList(self.player.concealedTiles)
            # FIXME: oder player.concealedTiles schon immer sortieren, oder desse Type auf TileList aendern
        logExposed.sort()
        physExposed.sort()
        logConcealed.sort()
        physConcealed.sort()
        assert logExposed == physExposed, (
            f'{self.player}: exposed: player {logExposed} != hand {physExposed}. '
            f'Check those:{set(logExposed) ^ set(physExposed)}')
        assert logConcealed == physConcealed, (
            f'{self.player}: concealed: player {logConcealed} != hand {physConcealed}')


class PlayingHandBoard(HandBoard):

    """a board showing the tiles a player holds"""

    def __bestFocusTile(self, adding:Optional[List[UITile]]=None) ->Optional[UITile]:
        """Where should we put focus?"""
        source = adding if adding else self[:]
        focusCandidates = [x for x in source if x.focusable and x.tile.isConcealed]
        if not focusCandidates:
            # happens if we just exposed a claimed meld
            focusCandidates = [x for x in self[:] if x.focusable and x.tile.isConcealed]
        if focusCandidates := sorted(focusCandidates, key=lambda x: x.xoffset):
            return focusCandidates[0]
        return None

    def sync(self, adding:Optional[List[UITile]]=None) ->None:
        """place all tiles in HandBoard"""
        allTiles = self[:]
        if adding:
            allTiles.extend(adding)
        self.placeTiles(allTiles)
        if _ := self.__bestFocusTile(adding):
            self.focusTile = _
        assert Internal.scene
        Internal.scene.handSelectorChanged(self)
        self.hasLogicalFocus = bool(adding)

    @Board.focusTile.setter  # type: ignore
    def focusTile(self, uiTile):
        Board.focusTile.fset(self, uiTile)
        assert self.player
        assert Internal.scene
        if self.player and Internal.scene.clientDialog:
            Internal.scene.clientDialog.focusTileChanged()

    def setEnabled(self, enabled:bool) ->None:
        """enable/disable this board"""
        if isAlive(self):
            # aborting a running game: the underlying C++ object might
            # already have been destroyed
            super().setEnabled(enabled)
            assert self.player
            assert self.player.game
            self.tileDragEnabled &= self.player == self.player.game.myself

    def dragMoveEvent(self, event:Optional['QGraphicsSceneDragDropEvent']) ->None:
        """only dragging to discard board should be possible"""
        if event:
            event.setAccepted(False)

    def newLowerMelds(self) ->MeldList:
        """a list of melds for the hand as it should look after sync"""
        assert self.player
        if self.player.concealedMelds:
            result = MeldList(self.player.concealedMelds)
        elif self.player.concealedTiles:
            content = Hand(self.player, unusedTiles=self.player.concealedTiles)
            result = MeldList(content.melds + content.bonusMelds)
        else:
            return MeldList()
        assert Internal.Preferences
        if not Internal.Preferences.rearrangeMelds:
            result = MeldList(Meld(x) for x in result.tiles())
            # one meld per tile
        result.sort()
        return result

    def discard(self, tile:Tile) ->None:
        """select the rightmost matching tileItem and move it
        to DiscardBoard"""
        if self.focusTile and self.focusTile.tile is tile:
            lastDiscard = self.focusTile
        else:
            # if an opponent player discards, we want to discard from the
            # right end of the hand# thus minimizing tile movement
            # within the hand
            lastDiscard = self[tile][-1]  # type:ignore[index]
        assert Internal.scene
        cast('PlayingScene', Internal.scene).discardBoard.discardTile(lastDiscard)
        for uiTile in self:
            uiTile.focusable = False

    def addUITile(self, uiTile:UITile) ->None:
        """add uiTile to this board"""
        super().addUITile(uiTile)
        assert self.player
        assert self.player.game
        if uiTile.isBonus and not self.player.game.isScoringGame():
            Sound.bonus()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import weakref
from random import Random
from typing import Type, Any, TYPE_CHECKING, Optional, Sequence
from typing import Iterable, Protocol, TypeVar
from typing import MutableSequence

from util import callers
from common import Debug

if TYPE_CHECKING:
    from game import Game

_T_co = TypeVar("_T_co", covariant=True)
_T = TypeVar("_T")

class SupportsLenAndGetItem(Protocol[_T_co]):
    """just for type checking"""
    def __len__(self) -> int: ...
    def __getitem__(self, __k: int) -> _T_co: ...


class CountRandomCalls:

    """a helper class for logging count of random calls"""

    def __init__(self, rnd:'CountingRandom', what:str) ->None:
        self.rnd = rnd
        self.what = what
        self.oldCount = CountingRandom.count

    def __enter__(self) ->'CountRandomCalls':
        return self

    def __exit__(self, exc_type:Type[Exception], exc_value:Exception, trback:str) ->None:
        if Debug.random:
            if self.rnd.game:
                self.rnd.game.debug(
                    f'{CountingRandom.count - self.oldCount} out of '
                    f'{CountingRandom.count} calls to random by {self.what} from {callers()}')


class CountingRandom(Random):

    """counts how often random() is called and prints debug info"""

    count = 0

    def __init__(self, game:'Game', value:Any=None) ->None:
        self._game = weakref.ref(game)
        Random.__init__(self, value)
        CountingRandom.count = 0

    @property
    def game(self) ->Optional['Game']:
        """hide the fact that game is a weakref"""
        return self._game()

    def random(self) ->float:
        """the central randomizator"""
        CountingRandom.count += 1
        return super().random()

    def seed(self, a:Any=None, version:int=2) ->None:
        CountingRandom.count = 0
        super().seed(a, version)
        if Debug.random:
            if self.game:
                self.game.debug(f'Random gets seed {a}')

    def randrange(self, start:int, stop:Optional[int]=None, step:int=1) ->int:
        with CountRandomCalls(self, f'randrange({start},{stop},step={step})'):
            return super().randrange(start, stop, step)

    def choice(self, seq:SupportsLenAndGetItem[_T]) ->_T:
        """Choose a random element from a non-empty sequence."""
        if len(seq) == 1:
            return seq[0]
        with CountRandomCalls(self, f'choice({seq})'):
            return super().choice(seq)

    def sample(self, population:Sequence[_T],
        k:int, *, counts:Optional[Iterable[int]]=None) ->list[_T]:
        """add debug output to sample. Chooses k unique random elements"""
        with CountRandomCalls(self, f'sample({population}, {k})'):
            return super().sample(population, k, counts=counts)

    def shuffle(self, x:MutableSequence[Any], random:Optional[Any]=None) ->None:
        """add debug output to shuffle. Shuffles list x in place."""
        with CountRandomCalls(self, f'shuffle({x})'):
            try:
                # Python 3.10 or earlier
                # pylint:disable=deprecated-argument,too-many-function-args
                super().shuffle(x, random)  # type:ignore[call-arg]
            except TypeError:
                # Python 3.11 or later
                super().shuffle(x)



# -*- coding: utf-8 -*-
"""
Copyright (C) 2010-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

partially based on C++ code from:
Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import TYPE_CHECKING, Optional, List, Any, Type, Tuple, Dict, Sequence

from qt import QComboBox, QTableView, QSizePolicy, QAbstractItemView
from qt import QTransform

from kde import KIcon

from log import i18n

if TYPE_CHECKING:
    from qt import QObject, QWidget, QPainter, QGraphicsItem
    from rule import Ruleset


class MJTableView(QTableView):

    """a QTableView with app specific defaults"""

    def __init__(self, parent:Optional['QWidget']=None) ->None:
        super().__init__(parent)
        if header := self.horizontalHeader():
            header.setStretchLastSection(True)
        self.setAlternatingRowColors(True)
        pol = QSizePolicy()
        pol.setHorizontalPolicy(QSizePolicy.Policy.Expanding)
        pol.setVerticalPolicy(QSizePolicy.Policy.Expanding)
        self.setSizePolicy(pol)
        if header := self.verticalHeader():
            header.hide()

    def initView(self) ->None:
        """set some app specific defaults"""
        self.selectRow(0)
        self.resizeColumnsToContents()
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)


class ListComboBox(QComboBox):

    """easy to use with a python list. The elements must have an
    attribute 'name'."""

    def __init__(self, items:List[Any], parent:Optional['QWidget']=None) ->None:
        super().__init__(parent)
        self.items = items

    @property
    def items(self) ->List[Any]:
        """combo box items"""
        return [self.itemData(idx) for idx in range(self.count())]

    @items.setter
    def items(self, items:List['Ruleset']) ->None:
        """combo box items"""
        self.clear()
        if items:
            for item in items:
                self.addItem(i18n(item.name), item)

    def findItem(self, search:Any) ->int:
        """return the index or -1 of not found """
        for idx, item in enumerate(self.items):
            if item == search:
                return idx
        return -1

    def names(self) ->List[str]:
        """a list with all item names"""
        return list(x.name for x in self.items)

    @property
    def current(self) ->Any:
        """current item"""
        return self.itemData(self.currentIndex())

    @current.setter
    def current(self, item:Any) ->None:
        """current item"""
        newIdx = self.findItem(item)
        if newIdx < 0:
            raise IndexError(f'{item.name} not found in ListComboBox')
        self.setCurrentIndex(newIdx)


class Painter:

    """a helper class for painting: saves/restores painter"""

    def __init__(self, painter:'QPainter') ->None:
        """painter is the painter to be saved/restored"""
        self.painter = painter

    def __enter__(self) ->'Painter':
        self.painter.save()
        return self

    def __exit__(self, exc_type:Type[Exception], exc_value:Exception, trback:Any) ->None:
        """now check time passed"""
        self.painter.restore()


class BlockSignals:

    """a helper class for temporary blocking of Qt signals"""

    def __init__(self, qobjects:Sequence['QObject']) ->None:
        self.qobjects = qobjects

    def __enter__(self) ->None:
        for obj in self.qobjects:
            obj.blockSignals(True)

    def __exit__(self, exc_type:Type[Exception], exc_value:Exception, trback:Any) ->None:
        for obj in self.qobjects:
            obj.blockSignals(False)


def decorateWindow(window:'QWidget', name:str='') ->None:
    """standard Kajongg window title and icon"""
    if name:
        window.setWindowTitle(f"{name} – {i18n('kajongg')}")
    else:
        window.setWindowTitle(i18n('Kajongg'))
    window.setWindowIcon(KIcon('kajongg'))


def rotateCenter(item:'QGraphicsItem', angle:float) ->None:
    """rotates a QGraphicsItem around its center
    rotateCenter and sceneRotation could be a mixin class but there are so many
    classes needing this. If and when more QGraphicsItem* classes are changed
    to QGraphicsObject, those could be moved to GraphicsObject(QGraphicsObject)"""
    center = item.boundingRect().center()
    centerX, centerY = center.x() * item.scale(), center.y() * item.scale()
    item.setTransform(QTransform().translate(
        centerX, centerY).rotate(angle).translate(-centerX, -centerY))

def sceneRotation(item:'QGraphicsItem') ->int:
    """the combined rotation of item and all parents in degrees: 0,90,180 or 270"""
    transform = item.sceneTransform()
    matrix_raw:Tuple[int, ...] = (
        round(transform.m11()),  # horizontal scaling factor
        round(transform.m12()),  # vertical shearing factor
        round(transform.m21()),  # horizontal shearing factor
        round(transform.m22()))  # vertical scaling factor
    matrix = tuple(1 if x > 0 else -1 if x < 0 else 0 for x in matrix_raw)
    rotations:Dict[Tuple[int, ...], int] = {
        (0, 0, 0, 0): 0,
        (1, 0, 0, 1): 0,
        (0, 1, -1, 0): 90,
        (1, 1, -1, 1): 90,
        (-1, 0, 0, -1): 180,
        (0, -1, 1, 0): 270,
        (1, -1, 1, 1): 270}
    if matrix not in rotations:
        raise ValueError(f'{item=} matrix unknown:{matrix=} {matrix_raw=}')
    return rotations[matrix]



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import traceback
import datetime
import weakref
import gc
from typing import List, Dict, Any, TYPE_CHECKING, Optional, Tuple, Union, Sequence, Generator, cast

from twisted.spread import pb
from twisted.internet.defer import Deferred

from log import logInfo, logDebug, logException, logFailure
from mi18n import i18nE
from message import Message
from common import Debug, ReprMixin, id4
from move import Move

if TYPE_CHECKING:
    from player import PlayingPlayer
    from client import Client
    from user import User
    from game import PlayingGame
    from twisted.python.failure import Failure
    from servertable import ServerTable, ServerGame


class Request(ReprMixin):

    """holds a Deferred and related attributes, used as part of a DeferredBlock"""

    def __init__(self, block:'DeferredBlock', deferred:Deferred, user:'User') ->None:
        self._block = weakref.ref(block)
        self.deferred = deferred
        self._user:weakref.ReferenceType['User'] = weakref.ref(user)
        self.answer:Optional[Message] = None
        self.args:Optional[Union[Tuple[Any, ...], List[Any]]] = None  # FIXME: unify or explain
        self.startTime = datetime.datetime.now()
        player = self.block.playerForUser(user)
        self._player = weakref.ref(player) if player else None

    @property
    def block(self) ->'DeferredBlock':
        """hide weakref"""
        result = self._block()
        assert result
        return result

    @property
    def user(self) ->'User':
        """hide weakref"""
        result = self._user()
        assert result
        return result

    @property
    def player(self) ->Optional['PlayingPlayer']:
        """hide weakref"""
        return self._player() if self._player else None

    def gotAnswer(self, rawAnswer:Any) ->None:
        """convert the wired answer into something more useful"""
        if isinstance(rawAnswer, tuple):
            answer = rawAnswer[0]
            if isinstance(rawAnswer[1], tuple):
                self.args = rawAnswer[1]
            else:
                self.args = [rawAnswer[1]]
        else:
            answer = rawAnswer
            self.args = None
        if answer in Message.defined:
            self.answer = Message.defined[answer]
        else:
            if Debug.deferredBlock:
                logDebug(f'Request {self} ignores {rawAnswer}')

    def age(self) ->int:
        """my age in full seconds"""
        return int((datetime.datetime.now() - self.startTime).total_seconds())

    def __str__(self) ->str:
        cmd = self.deferred.command  # type:ignore[attr-defined]
        if self.answer:
            answer = str(self.answer) # TODO: needed?
        else:
            answer = 'OPEN'
        result = ''
        if Debug.deferredBlock:
            result += f'_{id4(self):>4} '
        result += f'{cmd}->{self.user.__class__.__name__}({self.user.name:<10}): {answer}'
        if self.age():
            result += f' after {self.age()} sec'
        return result

    def prettyAnswer(self) ->str:
        """for debug output"""
        if self.answer:
            result = str(self.answer)
        else:
            result = 'OPEN'
        if self.args:
            result += f"({','.join(str(x) for x in self.args)})"
        return result

    def pretty(self) ->str:
        """for debug output"""
        result = ''
        if Debug.deferredBlock:
            result += f'_{id4(self):>4} '
        result += f'{self.deferred.command:<12}<-{self.user.__class__.__name__:>6}'  # type:ignore[attr-defined]
        result += f'({self.user.name:<10}): ANS={self.prettyAnswer()}'
        if self.age() > 0:
            result += f' after {self.age()} sec'
        return result


class DeferredBlock(ReprMixin):

    """holds a list of deferreds and waits for each of them individually,
    with each deferred having its own independent callbacks. Fires a
    'general' callback after all deferreds have returned.
    Usage: 1. define, 2. add requests, 3. set callback"""

    blocks : List['DeferredBlock'] = []
    blockWarned = False  # did we already warn about too many blocks?

    def __init__(self, table:'ServerTable', temp:bool=False, where:Optional[str]=None) ->None:
        dummy, dummy, function, dummy = traceback.extract_stack()[-2]
        self.outstanding = 0
        self.calledBy = function
        if not temp:
            self.garbageCollection()
        self.where = where
        self.table = table
        self.requests:List[Request] = []
        self.callbackMethod = None
        self.__callbackArgs:Optional[Tuple[Any,...]] = None
        self.completed = False
        if not temp:
            DeferredBlock.blocks.append(self)
            if not DeferredBlock.blockWarned:
                if len([x for x in DeferredBlock.blocks if x.table == table]) > 10:
                    DeferredBlock.blockWarned = True
                    logInfo(f'We have {len(DeferredBlock.blocks)} DBlocks:')
                    for block in DeferredBlock.blocks:
                        logInfo(str(block))

    def debugPrefix(self, dbgMarker:str='') ->str:
        """prefix for debug message"""
        return (f'T{self.table.tableid} B_{id4(self):>4} {self.calledBy[:15]:<15} '
                f'{dbgMarker:<3}(out={self.outstanding})')

    def debug(self, dbgMarker:str, msg:str) ->None:
        """standard debug format"""
        logDebug(' '.join([self.debugPrefix(dbgMarker), msg]))

    def __str__(self) ->str:
        req = '[' + ','.join(str(x) for x in self.requests) + ']'
        return (f"{self.debugPrefix()} {req} {int(self.outstanding)} "
                f"{'is completed' if self.completed else 'not completed'} {self.prettyCallback()}")

    def outstandingStr(self) ->str:
        """like __str__ but only with outstanding answers"""
        req = f"[{','.join(str(x) for x in self.requests if not x.answer)}]"
        return f"{self.calledBy} callback={self.prettyCallback()}:{req}"

    @staticmethod
    def garbageCollection() ->None:
        """delete completed blocks. Only to be called before
        inserting a new block. Assuming that block creation
        never overlaps."""
        for block in DeferredBlock.blocks[:]:
            if block.callbackMethod is None:
                try:
                    block.logBug(f'DBlock {str(block)} has no callback')
                finally:
                    # we do not want DoS for future games
                    DeferredBlock.blocks.remove(block)
            if block.completed:
                DeferredBlock.blocks.remove(block)
        if len(DeferredBlock.blocks) > 100:
            logDebug(
                f'We have {len(DeferredBlock.blocks)} DeferredBlocks, they must be leaking')
            for _ in (id4(x) for x in gc.get_objects() if x.__class__.__name__ == 'DeferredBlock'):
                print(f'DeferredBlock {_} left, allocated by {_.where}')


    def __addRequest(self, deferred:Deferred, user:'User', about:Optional['PlayingPlayer']) ->None:
        """add deferred for user to this block"""
        assert self.callbackMethod is None, 'AddRequest: already have callback defined'
        assert not self.completed, 'AddRequest: already completed'
        request = Request(self, deferred, user)
        self.requests.append(request)
        self.outstanding += 1
        deferred.addCallback(
            self.__gotAnswer,
            request).addErrback(
                self.__failed,
                request)
        if Debug.deferredBlock:
            notifying = ' notifying' if deferred.notifying else ''  # type:ignore[attr-defined]
            rqString = (f"_{id4(request):>4} {deferred.command}{notifying} "  # type:ignore[attr-defined]
                        f"{about.name if about else ''}->"
                        f"{user.__class__.__name__:>6}({user.name:<10})")
            self.debug(f'+:{len(self.requests)}', rqString)

    def removeRequest(self, request:Request) ->None:
        """we do not want this request anymore"""
        self.requests.remove(request)
        if not request.answer:
            self.outstanding -= 1
        if Debug.deferredBlock:
            self.debug(f'-:{int(self.outstanding)}', str(request)) # TODO: auch ohne?
        self.callbackIfDone()

    def callback(self, method:Any, *args:Any) ->None:
        """to be done after all users answered"""
        assert not self.completed, 'callback already completed'
        assert self.callbackMethod is None, 'callback: no method defined'
        self.callbackMethod = method
        self.__callbackArgs = args
        if Debug.deferredBlock:
            self.debug('CB', self.prettyCallback())
        self.callbackIfDone()

    def __gotAnswer(self, result:Deferred, request:Request) ->None:
        """got answer from user"""
        if request in self.requests:
            # after having lost connection to client, an answer could still be
            # in the pipe
            if result is None:
                if Debug.deferredBlock:
                    self.debug('IGN', request.pretty())
                return
            request.gotAnswer(result)
            assert request.answer
            if hasattr(request.user, 'pinged'):
                # a Client (for robots) does not have it
                request.user.pinged()
            if Debug.deferredBlock:
                self.debug('ANS', request.pretty())
            if hasattr(request.answer, 'notifyAction'):
                block = DeferredBlock(self.table, temp=True, where='__gotAnswer')
                if hasattr(request.answer, 'receivers'):
                    # if the request wants to send info to receivers:
                    receivers = request.answer.receivers(request)
                    if receivers:
                        block.tell(
                            request.player,
                            receivers,
                            request.answer,
                            notifying=True)
            self.outstanding -= 1
            assert self.outstanding >= 0, f'__gotAnswer: outstanding {int(self.outstanding)}'
            self.callbackIfDone()
        else:
            if Debug.deferredBlock:
                self.debug('NOP', request.pretty())

    def __failed(self, result:'Failure', request:Request) ->None:
        """a user did not or not correctly answer"""
        if request in self.requests:
            self.removeRequest(request)
        if result.type in [pb.PBConnectionLost]:
            msg = i18nE('The game server lost connection to player %1')
            self.table.abort(msg, request.user.name)
        else:
            msg = i18nE('Error for player %1: %2\n%3')
            if hasattr(result, 'traceback'):
                traceBack = result.traceback
            else:
                traceBack = result.getBriefTraceback()
            if not isinstance(result.value, AssertionError):
                self.table.abort(
                    msg,
                    request.user.name,
                    result.getErrorMessage(),
                    traceBack)

    def logBug(self, msg:str) ->None:
        """log msg and raise exception"""
        for request in self.requests:
            logDebug(str(request)) # TODO:
        logException(msg)

    def callbackIfDone(self) ->None:
        """if we are done, convert received answers to something more useful and callback"""
        if self.completed:
            return
        assert self.outstanding >= 0, f'callbackIfDone: outstanding {int(self.outstanding)}'
        if self.outstanding == 0 and self.callbackMethod is not None:
            self.completed = True
            if any(not x.answer for x in self.requests):
                self.logBug(
                    f'Block {str(self)}: Some requests are unanswered')
            if Debug.deferredBlock:
                commandText = []
                for command in sorted({x.deferred.command for x in self.requests}):
                    text = f'{command}:'
                    answerList = []
                    for answer in sorted({x.prettyAnswer() for x in self.requests if x.deferred.command == command}):
                        answerList.append((answer, [
                            x for x in self.requests
                            if x.deferred.command == command and answer == x.prettyAnswer()]))
                    answerList = sorted(answerList, key=lambda x: len(x[1]))
                    answerTexts = []
                    if len(answerList) == 1:
                        answerTexts.append(
                            f'{answerList[-1][0]} from all')
                    else:
                        for answer, requests in answerList[:-1]:
                            answerTexts.append(
                                f"{answer} from {','.join(x.user.name for x in requests)}")
                        answerTexts.append(
                            f'{answerList[-1][0]} from others')
                    text += ', '.join(answerTexts)
                    commandText.append(text)
                methodName = self.prettyCallback()
                if methodName:
                    methodName = f' next:{methodName}'
                self.debug(
                    'END',
                    f"{' / '.join(commandText)} {methodName}")
            if self.callbackMethod is not False:
                self.callbackMethod(self.requests, *self.__callbackArgs)

    def prettyCallback(self) ->str:
        """pretty string for callbackMethod"""
        if self.callbackMethod is False:
            result = ''
        elif self.callbackMethod is None:
            result = 'None'
        else:
            result = self.callbackMethod.__name__
            if self.__callbackArgs:
                result += f"({','.join([str(x) for x in self.__callbackArgs] if self.__callbackArgs else '')})"
        return result

    def playerForUser(self, user:'User') ->Optional['PlayingPlayer']:
        """return the game player matching user"""
        if user.__class__.__name__.endswith('Player'):
            assert False, f'playerForUser must get User, not {user}'
        if self.table.game:
            for player in self.table.game.players:
                if user.name == player.name:
                    return player
        return None

    @staticmethod
    def __enrichMessage(game:Optional['ServerGame'], about:Optional['PlayingPlayer'],
        command:Message, kwargs:Dict[Any,Any]) ->None:
        """add supplemental data for debugging"""
        if command.sendScore and about:
            # the clients will compare our status with theirs. This helps
            # very much in finding bugs.
            kwargs['score'] = str(about.hand)
        if game and game.gameid and 'token' not in kwargs:
            # this lets the client assert that the message is meant for the
            # current hand
            kwargs['token'] = game.point.token(game)
        else:
            kwargs['token'] = None

    def __convertReceivers(self, receivers:Sequence[Union[
        'PlayingPlayer', 'User']]) ->Generator[Union['User', 'Client'], None, None]:
        """try to convert Player to User or Client where possible"""
        for rec in receivers:
            if rec.__class__.__name__ == 'User':
                yield rec  # type:ignore[misc]
            else:
                yield self.table.remotes[rec]  # type:ignore[misc,index]

    def tell(self, about:Optional[Union['User', 'PlayingPlayer']],
        receivers:Sequence[Union['User', 'PlayingPlayer']], command:Message, **kwargs:Any) ->None:
        """send info about player 'about' to users 'receivers'"""
        def encodeKwargs() ->None:
            """those values are classes like Meld, Tile etc.
               Convert to bytes"""
            _ = tuple(kwargs.keys())
            for keyword in _:
                if any(keyword.lower().endswith(x) for x in ('tile', 'tiles', 'meld', 'melds')):
                    if kwargs[keyword] is not None:
                        kwargs[keyword] = str(kwargs[keyword])
                if keyword == 'players':
                    kwargs['playerNames'] = [(x.wind, x.name) for x in kwargs[keyword]]
                    del kwargs['players']
        encodeKwargs()
        if about.__class__.__name__ == 'User':
            aboutPlayer = self.playerForUser(about)  # type:ignore[arg-type]
        else:
            aboutPlayer = about  # type:ignore[assignment]
        assert isinstance(receivers, list), f'receivers should be list: {type(receivers)}/{repr(receivers)}'
        assert receivers, f'DeferredBlock.tell({command}) has no receiver'
        self.__enrichMessage(self.table.game, aboutPlayer, command, kwargs)
        aboutName = aboutPlayer.name if aboutPlayer else None
        if self.table.running and len(receivers) in [1, 4]:
            # messages are either identical for all 4 players
            # or identical for 3 players and different for 1 player. And
            # we want to capture each message exactly once.
            assert self.table.game
            self.table.game.moves.append(Move(aboutPlayer, command, kwargs))
        localDeferreds = []
        for rec in self.__convertReceivers(receivers):
            defer:Deferred
            isClient = rec.__class__.__name__.endswith('Client')
            if isClient:
                defer = Deferred()
                defer.addCallback(cast('Client', rec).remote_move, command, **kwargs).addErrback(logFailure)
                defer.command = command.name  # type:ignore[attr-defined]
                defer.notifying = 'notifying' in kwargs  # type:ignore[attr-defined]
                self.__addRequest(defer, rec, aboutPlayer)  # type:ignore[arg-type]
                localDeferreds.append(defer)
            else:
                if Debug.traffic:
                    message = (f"-> {rec.name[:15] if rec.name else 'NOBODY':<15} "
                               f"about {aboutPlayer} {command}{kwargs!r}")
                    logDebug(message)
                defer = self.table.server.callRemote(
                    rec,  # type:ignore[arg-type]
                    'move',
                    aboutName,
                    command.name,
                    **kwargs)
                if defer:
                    defer.command = command.name  # type:ignore[attr-defined]
                    defer.notifying = 'notifying' in kwargs  # type:ignore[attr-defined]
                    self.__addRequest(defer, rec, aboutPlayer)  # type:ignore[arg-type]
                else:
                    msg = i18nE('The game server lost connection to player %1')
                    self.table.abort(msg, rec.name)


        for defer in localDeferreds:
            defer.callback(aboutName)  # callback needs an argument !

    def tellPlayer(self, player:'PlayingPlayer', command:Message, **kwargs:Any) ->None:
        """address only one user"""
        self.tell(player, [player], command, **kwargs)

    def tellOthers(self, player:'PlayingPlayer', command:Message, **kwargs:Any) ->None:
        """tell others about player'"""
        assert self.table.game
        self.tell(
            player,
            list(
                x for x in self.table.game.players if x.name != player.name),
            command,
            **kwargs)

    def tellAll(self, player:Optional['PlayingPlayer'], command:Message, **kwargs:Any) ->None:
        """tell something to all players"""
        assert self.table.game
        self.tell(player, self.table.game.players, command, **kwargs)



# By Simon Edwards <simon@simonzone.com>
# This file is in the public domain.

"""
Byte-compiles a given Python source file, generating a .pyc file or, if the
Python executable was invoked with -O, a .pyo file from it.
It uses the --destination-dir option to set the path to the source file (which
will appear in tracebacks, for example), so that if the .py file was in a build
root will appear with the right path.
"""

import argparse
import os
import py_compile


if __name__ == '__main__':
    parser = argparse.ArgumentParser('Byte-compiles a Python source file.')
    parser.add_argument('-d', '--destination-dir', required=True,
                        help='Location where the source file will be '
                        'installed, without any build roots.')
    parser.add_argument('source_file',
                        help='Source file to byte-compile.')

    args = parser.parse_args()

    dfile = os.path.join(args.destination_dir,
                         os.path.basename(args.source_file))
    py_compile.compile(args.source_file, dfile=dfile)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import datetime

from typing import Optional, List, Any, TYPE_CHECKING, Union

from qt import Qt, QModelIndex, QPersistentModelIndex
from qt import QAbstractTableModel, QDialogButtonBox, QDialog
from qt import QHBoxLayout, QVBoxLayout, QCheckBox
from qt import QItemSelectionModel, QAbstractItemView

from dialogs import WarningYesNo
from kde import KIcon
from mi18n import i18n, i18nc
from query import Query
from guiutil import MJTableView, decorateWindow
from statesaver import StateSaver
from common import Debug
from modeltest import ModelTest

if TYPE_CHECKING:
    from qt import QWidget, QEvent, QKeyEvent
    from game import PlayingGame

class GamesModel(QAbstractTableModel):

    """data for the list of games"""

    def __init__(self) ->None:
        super().__init__()
        self._resultRows:List[List[Any]] = []

    def columnCount(self, unusedParent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """including the hidden col 0"""
        return 3

    def rowCount(self, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->int:
        """how many games"""
        if parent.isValid():
            # we have only top level items
            return 0
        return len(self._resultRows)

    def setResultset(self, rows:List[List[Any]]) ->None:
        """new data"""
        self.beginResetModel()
        try:
            self._resultRows = rows
        finally:
            self.endResetModel()

    def index(self, row:int, column:int, parent:Union[QModelIndex,QPersistentModelIndex]=QModelIndex()) ->QModelIndex:
        """helper"""
        if (row < 0
                or column < 0
                or row >= self.rowCount(parent)
                or column >= self.columnCount(parent)):
	    # similar in tree.py
            return QModelIndex()
        return self.createIndex(row, column, 0)

    def data(self, index:Union[QModelIndex,QPersistentModelIndex], role:int=Qt.ItemDataRole.DisplayRole) ->Any:
        """get score table from view"""
        assert isinstance(index, QModelIndex)
        if role is None:
            role = Qt.ItemDataRole.DisplayRole
        if not (index.isValid() and role == Qt.ItemDataRole.DisplayRole):
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            unformatted = str(
                self._resultRows[index.row()][index.column()]) # TODO: brauche ich str?
            if index.column() == 2:
                # we do not yet use this for listing remote games but if we do
                # this translation is needed for robot players
                names = [i18n(name) for name in unformatted.split('///')]
                return ', '.join(names)
            if index.column() == 1:
                dateVal = datetime.datetime.strptime(
                    unformatted, '%Y-%m-%dT%H:%M:%S')
                return dateVal.strftime('%c')
            if index.column() == 0:
                return int(unformatted)
        return super().data(index, role)

    def headerData(self, section:int, orientation:Qt.Orientation, role:int=Qt.ItemDataRole.DisplayRole) ->Optional[str]:
        """for the two visible columns"""
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return i18n('Players') if section == 2 else i18n('Started')
        return None


class Games(QDialog):

    """a dialog for selecting a game"""

    def __init__(self, parent:Optional['QWidget']=None) ->None:
        super().__init__(parent)
        self.selectedGame = 0
        self.onlyPending = True
        decorateWindow(self, i18nc("@title:window", "Games"))
        self.setObjectName('Games')
        self.resize(700, 400)
        self.model = GamesModel()
        if Debug.modelTest:
            self.modelTest = ModelTest(self.model, self)

        self.view = MJTableView(self)
        self.view.setModel(self.model)
        self.selection = QItemSelectionModel(self.model, self.view)
        self.view.setSelectionModel(self.selection)
        self.view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)

        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel)
        self.newButton = self.buttonBox.addButton(
            i18nc('start a new game', "&New"), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.newButton
        self.newButton.setIcon(KIcon("document-new"))
        self.newButton.clicked.connect(self.accept)
        self.loadButton = self.buttonBox.addButton(
            i18n("&Load"), QDialogButtonBox.ButtonRole.AcceptRole)
        assert self.loadButton
        self.loadButton.clicked.connect(self.loadGame)
        self.loadButton.setIcon(KIcon("document-open"))
        self.deleteButton = self.buttonBox.addButton(
            i18n("&Delete"), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.deleteButton
        self.deleteButton.setIcon(KIcon("edit-delete"))
        self.deleteButton.clicked.connect(self.delete)

        chkPending = QCheckBox(i18n("Show only pending games"), self)
        chkPending.setChecked(True)
        cmdLayout = QHBoxLayout()
        cmdLayout.addWidget(chkPending)
        cmdLayout.addWidget(self.buttonBox)

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.addLayout(cmdLayout)
        self.setLayout(layout)
        StateSaver(self)

        self.selection.selectionChanged.connect(self.selectionChanged)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.view.doubleClicked.connect(self.loadGame)
        chkPending.stateChanged.connect(self.pendingOrNot)

    def showEvent(self, unusedEvent:Optional['QEvent']) ->None:
        """only now get the data set. Not doing this in__init__ would eventually
        make it easier to subclass from some generic TableEditor class"""
        self.setQuery()
        self.view.initView()
        self.selectionChanged()

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """use insert/delete keys for insert/delete"""
        if event:
            key = event.key()
            if key == Qt.Key.Key_Delete:
                self.delete()
                event.ignore()
                return
            super().keyPressEvent(event)

    def selectionChanged(self) ->None:
        """update button states according to selection"""
        selectedRows = len(self.selection.selectedRows())
        if button := self.loadButton:
            button.setEnabled(selectedRows == 1)
        if button := self.deleteButton:
            button.setEnabled(selectedRows >= 1)

    def setQuery(self) ->None:
        """define the query depending on self.OnlyPending"""
        query = Query(
            "select g.id, g.starttime, " # pylint:disable=consider-using-f-string
            "p0.name||'///'||p1.name||'///'||p2.name||'///'||p3.name "
            "from game g, player p0,"
            "player p1, player p2, player p3 "
            "where seed=0"
            " and p0.id=g.p0 and p1.id=g.p1 "
            " and p2.id=g.p2 and p3.id=g.p3 "
            "%s"
            "and exists(select 1 from score where game=g.id)" %
            ("and g.endtime is null " if self.onlyPending else ""))
        self.model.setResultset(query.records)
        self.view.hideColumn(0)

    def __idxForGame(self, game:int) ->QModelIndex:
        """return the model index for game"""
        for row in range(self.model.rowCount()):
            idx = self.model.index(row, 0)
            if self.model.data(idx, 0) == game:
                return idx
        return self.model.index(0, 0)

    def __getSelectedGame(self) ->int:
        """return the game id of the selected game"""
        rows = self.selection.selectedRows()
        if rows:
            return self.model.data(rows[0], 0)
        return 0

    def pendingOrNot(self, chosen:Qt.CheckState) ->None:
        """do we want to see all games or only pending games?"""
        if self.onlyPending != bool(chosen):
            self.onlyPending = bool(chosen)
            prevSelected = self.__getSelectedGame()
            self.setQuery()
            if prevSelected:
                idx = self.__idxForGame(prevSelected)
                self.view.selectRow(idx.row())
        self.view.setFocus()

    def loadGame(self) ->None:
        """load a game"""
        self.selectedGame = self.__getSelectedGame()
        self.buttonBox.accepted.emit()

    def delete(self) ->None:
        """delete a game"""
        def answered(result:Any, games:List[int]) ->None:
            """question answered, result is True or False"""
            if result is True:
                for game in games:
                    Query("DELETE FROM score WHERE game = ?", (game, ))
                    Query("DELETE FROM game WHERE id = ?", (game, ))
                self.setQuery()  # just reload entire table
        if model := self.view.selectionModel():
            allGames = model.selectedRows(0)
            deleteGames = [x.data() for x in allGames]
            if not deleteGames:
                return
        else:
            return
        WarningYesNo(
            i18n(
                "Do you really want to delete <numid>%1</numid> games?<br>"
                "This will be final, you cannot cancel it with "
                "the cancel button",
                len(deleteGames))).addBoth(answered, deleteGames)



# -*- coding: utf-8 -*-

"""
Authors of original libkmahjongg in C++:
    Copyright (C) 1997 Mathias Mueller <in5y158@public.uni-hamburg.de>
    Copyright (C) 2006 Mauricio Piacentini <mauricio@tabuleiro.com>

this adapted python code:
    Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Optional, List, Tuple

from qt import QSizeF, QSvgRenderer
from log import logException, i18n
from mjresource import Resource

from common import LIGHTSOURCES, Internal, ReprMixin


class Tileset(Resource, ReprMixin):

    """represents a complete tileset"""

    resourceName : Optional[str] = 'tileset'
    configGroupName : str = 'KMahjonggTileset'

    def __init__(self, name:Optional[str]=None) ->None:
        """continue __build"""
        super().__init__(name)
        self.tileSize:QSizeF
        self.faceSize:QSizeF
        self.__shadowOffsets:List[List[Tuple[int,int]]]
        self.darkenerAlpha = 120 if self.desktopFileName == 'jade' else 50

        graphName = str(self.group.readEntry("FileName"))
        self.graphicsPath = Tileset.locate(graphName)
        if not self.graphicsPath:
            logException(
                f'cannot find kmahjongglib/tilesets/{graphName} for {self.desktopFileName}')
        self.renderer = QSvgRenderer(self.graphicsPath)
        if not self.renderer.isValid():
            logException(
                i18n(
                    'file <filename>%1</filename> contains no valid SVG',
                self.graphicsPath))
        distance = 0
        if self.desktopFileName == 'classic':
            distance = 2
        distanceSize = QSizeF(distance, distance)
        self.faceSize = self.renderer.boundsOnElement('BAMBOO_1').size()
        self.faceSize += distanceSize
        self.tileSize = self.renderer.boundsOnElement('TILE_2').size()
        self.tileSize += distanceSize
        shW = self.shadowWidth()
        shH = self.shadowHeight()
        self.__shadowOffsets = [
            [(-shW, 0), (0, 0), (0, shH), (-shH, shW)],
            [(0, 0), (shH, 0), (shW, shH), (0, shW)],
            [(0, -shH), (shH, -shW), (shW, 0), (0, 0)],
            [(-shW, -shH), (0, -shW), (0, 0), (-shH, 0)]]

        self.svgName = {
            'wn': 'WIND_1', 'ws': 'WIND_2', 'we': 'WIND_3', 'ww': 'WIND_4',
            'db': 'DRAGON_1', 'dg': 'DRAGON_2', 'dr': 'DRAGON_3'}
        for value in '123456789':
            self.svgName[f's{value}'] = f'ROD_{value}'
            self.svgName[f'b{value}'] = f'BAMBOO_{value}'
            self.svgName[f'c{value}'] = f'CHARACTER_{value}'
        for idx, wind in enumerate('eswn'):
            self.svgName[f'f{wind}'] = f'FLOWER_{int(idx + 1)}'
            self.svgName[f'y{wind}'] = f'SEASON_{int(idx + 1)}'

    def __str__(self) ->str:
        return self.desktopFileName

    @staticmethod
    def current() ->'Tileset':
        """the currently wanted tileset. If not yet defined, do so"""
        assert Internal.Preferences
        return Tileset(str(Internal.Preferences.tilesetName))

    def shadowWidth(self) ->int:
        """the size of border plus shadow"""
        return int(self.tileSize.width() - self.faceSize.width())

    def shadowHeight(self) ->int:
        """the size of border plus shadow"""
        return int(self.tileSize.height() - self.faceSize.height())

    def shadowOffsets(self, lightSource:str, rotation:int) ->Tuple[int, int]:
        """real offset of the shadow on the screen"""
        assert Internal.Preferences
        if not Internal.Preferences.showShadows:
            return (0, 0)
        lightSourceIndex = LIGHTSOURCES.index(lightSource)
        return self.__shadowOffsets[lightSourceIndex][rotation // 90]

    def tileFaceRelation(self) ->Tuple[float,float]:
        """return how much bigger the tile is than the face"""
        return (self.tileSize.width() / self.faceSize.width(),
                self.tileSize.height() / self.faceSize.height())



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from functools import total_ordering
import sys

from typing import Optional, Union, Tuple, Any, TYPE_CHECKING

from common import ReprMixin
from wind import Wind, East
from query import Query
from log import logError

if TYPE_CHECKING:
    from game import Game


@total_ordering
class Point(ReprMixin):

    """A point in time of a game. Useful for positioning after abort/restart

        Point does not care about wind or players or about any rules.

        seed: the seed for the random generator, shown in window title
        prevailing: the prevailing wind of the current round
        rotated: how often we rotated within current round
        notRotated: how often we did NOT rotate since last rotation
        moveCount: Number of executed moves within current hand
    """

    def __init__(self, source:Any) -> None:
        """Default values point to start of game"""
        self.seed = 0
        self.prevailing:Wind = East
        self.rotated = 0
        self.notRotated = 0
        self.handCount = 0   # unique ID over all hands of a game for data base
        self.moveCount = 0   # within current hand

        if isinstance(source, str):
            self.__init_from_string(source)
        elif isinstance(source, int):
            self.__init_from_db(source)
        elif isinstance(source, Point):
            self.__init_from_point(source)
        elif source.__class__.__name__.startswith('Query_'):
            self.__init_from_query(source)
        else:
            raise ValueError(f'Point() does not accept source {source}')

    def __init_from_string(self, string:str) ->None:
        """Init myself from string

            String format: SEED/Wrnm
            where only trailing parameters can be omitted.

            SEED: 1..n digits: the seed
            W: 1 char:         prevailing wind
            r: 0..1 digit:     rotations within current round
            n: 0..n chars:     how often we currently did not rotate. Encoded in letters a..z
            c: 0..n digits:    handCount
            m: 0..n chars:     Number of current move within current hand. Encoded in letters a..z
        """

        self.seed = int(string.split('/')[0])
        normalized = string
        if normalized.endswith('/'):
            normalized = normalized[:-1]
        if '/' in normalized:
            rest = normalized.split('/')[1]
            self.prevailing = Wind(rest[0])
            rest = rest[1:]
            if rest:
                self.rotated = int(rest[0])
                rest = rest[1:]
                self.notRotated, rest = self.decode_from_az(rest)
                if rest:
                    digits = ''.join(filter(str.isdigit, rest))
                    self.handCount = int(digits)
                    self.moveCount, _ = self.decode_from_az(rest[len(digits):])

    def __init_from_point(self, other:'Point') ->None:
        """Init myself from a Game instance"""
        self.seed = other.seed
        self.prevailing = other.prevailing
        self.rotated = other.rotated
        self.notRotated = other.notRotated
        self.moveCount = other.moveCount
        self.handCount = other.handCount

    def __init_from_db(self, gameid:int) ->None:
        """last recorded position"""
        self.handCount, self.rotated, self.notRotated, self.prevailing = Query(
            'select {fields} from score where game=? order by hand desc limit 1', (gameid, ),
            fields='hand, rotated, notrotated, prevailing').tuple()

    def __init_from_query(self, query:Any) ->None:
        """last recorded position"""
        self.seed = query.game
        self.prevailing = query.prevailing
        self.rotated = query.rotated
        self.notRotated = query.notrotated
        self.handCount = query.hand

    def is_in_first_hand(self) ->bool:
        """True only if this point is in the first hand of a game"""
        return self.handCount == 0

    def endHand(self) ->None:
        """hand is done"""
        self.handCount += 1
        self.notRotated += 1  # if we rotate, we will call Point.rotate()

    def rotate(self) ->None:
        """rotate winds"""
        self.notRotated = 0
        self.rotated += 1
        if self.rotated == 4:
            self.prevailing = next(self.prevailing)
            self.rotated = 0

    def prompt(self, game:'Game', withSeed:bool=True, withAI:bool=True, withMoveCount:bool=False) ->str:
        """
        Identifies the hand for window title and scoring table.

        @param withSeed: If set, include the seed used for the
        random generator.
        @type  withSeed: C{Boolean}
        @param withAI:   If set and AI != DefaultAI: include AI name for
        human players.
        @type  withAI:   C{Boolean}
        @param withMoveCount:   If set, include the current count of moves.
        @type  withMoveCount:   C{Boolean}
        @return:         The prompt.
        @rtype:          C{str}
        """
        aiVariant = ''
        if withAI and game and game.belongsToHumanPlayer():
            if game.myself:
                aiName = game.myself.intelligence.name()
            else:
                aiName = 'DefaultAI'
            if aiName != 'DefaultAI':
                aiVariant = aiName + '/'
        if withSeed:
            seedStr = str(self.seed)
        else:
            seedStr = ''
        delim = '/' if withSeed or withAI else ''
        _ = self.notRotated_as_str() or ' '
        result = f'{aiVariant}{seedStr}{delim}{self.prevailing}{self.rotated + 1}{_}'
        if withMoveCount:
            result += f'/{int(self.moveCount):3}'
        return result

    @property
    def roundsFinished(self) -> int:
        """wind index"""
        return self.prevailing.__index__()

    def seedFactor(self) -> int:
        """we want a predictable but different seed for every starting point"""
        return ((self.roundsFinished + 1) * 10000
               + self.rotated * 1000
               + self.notRotated * 100)

    def encode_to_az(self, value:int) ->str:
        """encode an integer into string"""
        result = ''
        num = value
        while num:
            result = chr(ord('a') + (num - 1) % 26) + result
            num = (num - 1) // 26
        return result

    def decode_from_az(self, value:str) ->Tuple[int, str]:
        """decode and return rest"""
        rest = value[:]
        result = 0
        while rest and rest[0] >= 'a' and rest[0] <= 'z':
            result = result * 26 + ord(rest[0]) - ord('a') + 1
            rest = rest[1:]
        return result, rest

    def notRotated_as_str(self) ->str:
        """encode into a..z"""
        return self.encode_to_az(self.notRotated)

    def token(self, game:'Game') ->str:
        """server and client use this for checking whether they talk about
        the same thing"""
        return self.prompt(game, withAI=False)

    def full_str(self) ->str:
        """with all attributes, opposite to __init_from_string"""
        return f'{self}{self.handCount}{self.encode_to_az(self.moveCount)}'

    def __str__(self) ->str:
        return f'{self.prevailing}{self.rotated}{self.notRotated_as_str()}'

    def __eq__(self, other:object) ->bool:
        if other is None:
            return False
        if not isinstance(other, Point):
            return NotImplemented
        return (self.prevailing, self.rotated, self.notRotated) == \
                (other.prevailing, other.rotated, other.notRotated)

    def __ne__(self, other:object) ->bool:
        return not self == other

    def __gt__(self, other:object) ->bool:
        if other is None:
            # open end
            return False
        if not isinstance(other, Point):
            return NotImplemented
        return (self.prevailing, self.rotated, self.notRotated) > (
            other.prevailing, other.rotated, other.notRotated)

    def __lt__(self, other:object) ->bool:
        if other is None:
            # open end
            return True
        if not isinstance(other, Point):
            return NotImplemented
        return (self.prevailing, self.rotated, self.notRotated) < (
            other.prevailing, other.rotated, other.notRotated)

class PointRange(ReprMixin):

    """Represents a range of points: Point..Point
        start and end are included.
        If end is None: open end
    """

    def __init__(self, game:'Game',
            first_point:Optional[Point]=None,
            last_point:Optional[Point]=None) ->None:

        self.first_point: Point
        self.last_point: Union[Point, None]  # None means open end

        if first_point is None and last_point is None:
            self.first_point = game.first_point
            self.last_point = game.last_point
        else:
            assert first_point
            self.first_point = first_point
            self.last_point = last_point

        if self.first_point > self.last_point:
            raise UserWarning(f'{first_point}..{last_point} is a negative range')

    @classmethod
    def from_string(cls, string:str) -> Tuple[Point, Union[Point, None]]:
        """parse a string like first..last"""
        if string is None:
            string = ''

        full_str = string
        if '/' not in full_str:
            full_str += '/'
        seed_str, pos_str = full_str.split('/')
        try:
            seed = int(seed_str)
        except ValueError:
            logError(f'Cannot parse seed from {full_str}, I need something like 15/E0..W4', showStack=False)
            sys.exit(1)

        positions = pos_str.split('..')
        first_pos = Point(f'{seed}/{positions[0]}')

        last_pos:Optional[Point] = None   # default
        if len(positions) > 1:
            last_pos = Point(f'{seed}/{positions[1]}')

        if first_pos > last_pos:
            logError(f'start is after end:{full_str}', showStack=False)
            sys.exit(1)
        return first_pos, last_pos

    @property
    def seed(self) ->int:
        """from first_pos"""
        return self.first_point.seed

    def __str__(self) ->str:
        """used in traffic"""
        result = f'{self.first_point.seed}/{self.first_point}'
        if self.last_point:
            result += f'..{self.last_point}'
        return result
